"""
CPU EMULATOR BUILTIN FUNCTIONS

These functions are used to emulate the effects of known builtin functions.

Add any builtin functions that need to be handled below.  The function should be declared as such

# Using the same function for multiple instructions:
@builtin_func("memmove")
@builtin_func("memcpy")
def _memcpy(cpu_context, call_ip, func_name, func_args):
    print "IN memmove or memcpy"
    return 1  # Return anything to be placed into rax (or equivalent)

# Using a single function for a builtin
@builtin_func
def memmove(cpu_context, call_ip, func_name, func_args):
    print "IN memmove"

"""

import six

import logging
from contextlib import contextmanager

from . import utils
from .constants import *
from .registry import registrar

logger = logging.getLogger(__name__)


# Dictionary containing builtin function names -> function
BUILTINS = {}
builtin_func = registrar(BUILTINS, name="builtin")

# Collections of user defined hooks.
_USER_DEFINED = None


def get(func_name_or_start_ea):
    """
    Gets function hook for given function name or start address.

    :param func_name_or_start_ea: Name or start address of the function to get hook for.

    :return: A function to run.
    :rtype: function
    """
    # Convert the string to lowercase so it can match the dictionary of built-in functions
    if isinstance(func_name_or_start_ea, str):
        func_name_or_start_ea = func_name_or_start_ea.lower()

    # First check user defined hooks.
    if _USER_DEFINED and func_name_or_start_ea in _USER_DEFINED:
        return _USER_DEFINED[func_name_or_start_ea]

    # Then pull from our builtins.
    return BUILTINS.get(func_name_or_start_ea)


@contextmanager
def hooks(hooks):
    """
    Context manager used to temporarily set the given user defined function hooks
    during emulation.

    :param hooks: Dictionary of function names mapping to hook functions.
    :return:
    """
    global _USER_DEFINED
    try:
        _USER_DEFINED = hooks
        yield
    finally:
        _USER_DEFINED = None


@builtin_func("_alloca")
@builtin_func("_malloca")
def _alloca(cpu_context, func_name, func_args):
    """
    Allocates stack space

    NOTE: Our memory controller will automatically allocate pages as it writes.
    Therefore, we don't actually need to allocate anything, just provide them with
    a pointer.
    """
    (size,) = func_args
    if size:
        size = utils.align_page_up(size)
        logger.debug(" :: Allocating {} bytes of stack space.".format(size))
        # Returns a pointer to our allocated stack space.
        return cpu_context.sp - size


@builtin_func("malloc")
@builtin_func("calloc")
def malloc(cpu_context, func_name, func_args):
    """
    Allocates memory block
    """
    if func_name == "calloc":
        num, size = func_args
        size = num * size
    else:
        (size,) = func_args
    if size:
        logger.debug(" :: Allocating {} bytes of memory.".format(size))
        return cpu_context.mem_alloc(size)


@builtin_func
def new(cpu_context, func_name, func_args):
    """
    Allocates memory block
    """
    size = func_args[0]
    # If we have a second argument it's either a nothrow_value or a ptr to where they
    # want us to allocate the space.
    # Just assume its a ptr if it's not 0 since the emulator should not produce a nothrow_value correctly
    # and therefore be 0.
    # TODO: Perhaps provide more context to make a more informed decision.
    try:
        ptr = func_args[1]
    except IndexError:
        ptr = 0
    if size:
        if ptr:
            logger.debug(" :: Mapping {} bytes into 0x{:X}".format(size, ptr))
            # If a pointer was provided, just give it back to them without doing anything.
            # Our memory controller will automatically allocate pages as they are being written.
            return ptr
        else:
            logger.debug(" :: Allocating {} bytes of memory.".format(size))
            return cpu_context.mem_alloc(size)


@builtin_func
def realloc(cpu_context, func_name, func_args):
    """
    Reallocates memory block
    """
    ptr, size = func_args
    if ptr and size:
        logger.debug(" :: Reallocating 0x{:X} with {} bytes.".format(ptr, size))
        return cpu_context.mem_realloc(ptr, size)


@builtin_func
def memchr(cpu_context, func_name, func_args):
    """
    Locate character in block of memory.

    Searches within the first num bytes of the block of memory pointed by data_ptr
    for the first occurrence of value (interpreted as an unsigned char) and returns a pointer
    to it.
    """
    data_ptr, value, num = func_args
    if six.PY2:
        value = chr(value)

    ptr = cpu_context.memory.find(value, start=data_ptr, end=data_ptr + num)
    if ptr == -1:
        return 0
    else:
        return ptr


@builtin_func("memmove")
@builtin_func("memmove_s")
@builtin_func("wmemmove")
@builtin_func("wmemmove_s")
@builtin_func("memcpy")
@builtin_func("memcpy_s")
@builtin_func("wmemcpy")
@builtin_func("wmemcpy_s")
def memcpy(cpu_context, func_name, func_args):
    """
    Copies count characters from src to dst.
    """
    secure = func_name.endswith("_s")
    wide = func_name.startswith("w")
    if secure:
        dst, dst_size, src, count = func_args
        count = min(dst_size, count)
    else:
        dst, src, count = func_args

    if wide:
        count *= 2

    if dst and src:
        logger.debug(" :: Copying {} bytes from 0x{:X} to 0x{:X}".format(count, src, dst))
        cpu_context.mem_copy(src, dst, count)
        return 0 if secure else dst


@builtin_func
def memset(cpu_context, func_name, func_args):
    """
    Writes count number of ch characters to dst.
    """
    dst, ch, count = func_args
    if dst:
        logger.debug(" :: Writing {!r} * {} to 0x{:X}".format(chr(ch), count, dst))
        cpu_context.mem_write(dst, chr(ch) * count)
    return dst


@builtin_func
def memcmp(cpu_context, func_name, func_args):
    lhs, rhs, count = func_args
    if lhs and rhs:
        logger.debug(" :: Comparing the first {} bytes in 0x{:X} with 0x{:X}".format(count, lhs, rhs))
        left = cpu_context.mem_read(lhs, count)
        right = cpu_context.mem_read(rhs, count)
        if left < right:
            return -1
        elif left > right:
            return 1
        else:
            return 0


@builtin_func("strcat")
@builtin_func("strcat_s")
@builtin_func("wcscat")
@builtin_func("wcscat_s")
@builtin_func("lstrcatA")
@builtin_func("lstrcatW")
def strcat(cpu_context, func_name, func_args):
    """
    Concatenates at most count characters from src, stopping if the null character
    is found, to the end of the null-terminated byte string pointed to by dst.
    The src[0] replaces the null terminator at the end of dst.
    """
    secure = func_name.endswith("_s")
    wide = func_name.startswith("w") or func_name.endswith("W")
    if secure:
        dst, dst_size, src = func_args
        if wide:
            dst_size *= 2
    else:
        dst, src = func_args
        dst_size = None

    if dst and src:
        logger.debug(" :: Concatenating c string in 0x{:X} to c string in 0x{:X}".format(src, dst))
        append_str = cpu_context.read_data(src, data_type=WIDE_STRING if wide else STRING)
        dest_str = cpu_context.read_data(dst, data_type=WIDE_STRING if wide else STRING)
        null_offset = dst + len(dest_str)
        if dst_size is not None:
            dst_size -= null_offset - dst
            append_str = append_str[:dst_size]
        terminator = b"\0\0" if wide else b"\0"
        cpu_context.mem_write(null_offset, append_str + terminator)
        return 0 if secure else dst


@builtin_func("strncat")
@builtin_func("strncat_s")
@builtin_func("wcsncat")
@builtin_func("wcsncat_s")
def strncat(cpu_context, func_name, func_args):
    """
    Concatenates at most count characters from src, stopping if the null character
    is found, to the end of the null-terminated byte string pointed to by dst.
    The src[0] replaces the null terminator at the end of dst.
    """
    secure = func_name.endswith("_s")
    wide = func_name.startswith("w")
    if secure:
        dst, dst_size, src, count = func_args
        if wide:
            dst_size *= 2
    else:
        dst, src, count = func_args
        dst_size = None
    if wide:
        count *= 2

    if dst and src:
        logger.debug(" :: Concatenating c string in 0x{:X} to c string in 0x{:X}".format(src, dst))
        append_str = cpu_context.read_data(src, data_type=WIDE_STRING if wide else STRING)[:count]
        dest_str = cpu_context.read_data(dst, data_type=WIDE_STRING if wide else STRING)
        null_offset = dst + len(dest_str)
        if dst_size is not None:
            dst_size -= null_offset - dst
            append_str = append_str[:dst_size]
        terminator = b"\0\0" if wide else b"\0"
        cpu_context.mem_write(null_offset, append_str + terminator)
        return 0 if secure else dst


@builtin_func
def strpbrk(cpu_context, func_name, func_args):
    """
    Locate characters in string.

    Returns a pointer to the first occurrence of str1 of any of the characters
    that are part of str2, or a null pointer if there are no matches.
    """
    str1_ptr, str2_ptr = func_args
    str1 = cpu_context.read_data(str1_ptr)
    str2 = cpu_context.read_data(str2_ptr)

    for offset, ch in enumerate(str1):
        if ch in str2:
            return str1_ptr + offset
    return 0


@builtin_func("strchr")
@builtin_func("strrchr")
def strchr(cpu_context, func_name, func_args):
    """
    Locate first or last occurrence of character in string.
    """
    string_ptr, character = func_args
    string = cpu_context.read_data(string_ptr)
    if six.PY2:
        character = chr(character)

    if func_name == "strchr":
        offset = string.find(character)
    else:
        offset = string.rfind(character)

    if offset == -1:
        return 0
    else:
        return string_ptr + offset


@builtin_func("strcpy")
@builtin_func("strcpy_s")
@builtin_func("wcscpy")
@builtin_func("wcscpy_s")
@builtin_func("lstrcpyA")
@builtin_func("lstrcpyW")
def strcpy(cpu_context, func_name, func_args):
    """
    Copies the null-terminated byte string pointed to by src, including the null terminator, to dst.
    """
    secure = func_name.endswith("_s")
    wide = func_name.startswith("w") or func_name.endswith("W")
    if secure:
        dst, dst_size, src = func_args
        if wide:
            dst_size *= 2
    else:
        dst, src = func_args
        dst_size = None

    if dst and src:
        logger.debug(" :: Copying c string in 0x{:X} to 0x{:X}".format(src, dst))
        terminator = b"\0\0" if wide else b"\0"
        src_str = cpu_context.read_data(src, data_type=WIDE_STRING if wide else STRING) + terminator
        size = len(src_str)
        if dst_size is not None:
            size = min(dst_size, size)  # limit to dst_size if secure.
        cpu_context.mem_copy(src, dst, size)
        return 0 if secure else dst


@builtin_func("strncpy")
@builtin_func("strncpy_s")
@builtin_func("wcsncpy")
@builtin_func("wcsncpy_s")
@builtin_func("lstrcpynA")
@builtin_func("lstrcpynW")
def strncpy(cpu_context, func_name, func_args):
    """
    Copies at most count characters of src (including the terminating null character) to dst
    but not any of the characters that follow the null character.
    For non-secure versions, if count is not reached, additional null characters are written to dst.
    """
    secure = func_name.endswith("_s")
    wide = func_name.startswith("w") or func_name.endswith("W")
    if secure:
        # Secure version has an extra dst_size argument used for validation, but we don't care to
        # throw errors if we don't have to!
        dst, _, src, count = func_args
    else:
        dst, src, count = func_args
    if wide:
        count *= 2

    if dst and src:
        logger.debug(" :: Copying c string in 0x{:X} to 0x{:X}".format(src, dst))
        terminator = b"\0\0" if wide else b"\0"
        src_str = cpu_context.read_data(src, data_type=WIDE_STRING if wide else STRING) + terminator
        size = min(count, len(src_str))
        cpu_context.mem_copy(src, dst, size)
        # Non-secure version also pads the rest with null characters to reach count size
        if not secure and size < count:
            # As a safety check, we shouldn't be writing more than about 0x1000 of padding.
            delta = count - size
            if delta > 0x1000:
                logger.warning(
                    " :: Attempted to write {} bytes of padding. Ignoring request and using {} bytes "
                    "of padding instead.".format(delta, 0x1000)
                )
                delta = 0x1000
            cpu_context.mem_write(src + size, b"\0" * delta)
        return 0 if secure else dst


@builtin_func
def strdup(cpu_context, func_name, func_args):
    """
    Returns a pointer to a null-terminated byte string, which is a duplicate of the string
    pointing to by str.
    """
    (str_ptr,) = func_args
    if str_ptr:
        logger.debug(" :: Copying c string in 0x{:X} to new pointer.".format(str_ptr))
        null_offset = cpu_context.memory.find(b"\0", start=str_ptr)
        size = null_offset - str_ptr
        # create new pointer
        new_ptr = cpu_context.mem_alloc(size + 1)
        # Copy at most size bytes of data then add the null terminator.
        cpu_context.mem_copy(str_ptr, new_ptr, size)
        cpu_context.mem_write(new_ptr + size, b"\0")
        return new_ptr


@builtin_func
def strndup(cpu_context, func_name, func_args):
    """
    Returns a pointer to a null-terminated byte string, which contains copies
    of at most size bytes from the string pointed to by str.
    """
    str_ptr, size = func_args
    if str_ptr:
        logger.debug(" :: Copying {} bytes of c string in 0x{:X} to new pointer.".format(size, str_ptr))
        null_offset = cpu_context.memory.find(b"\0", start=str_ptr)
        size - min(size, null_offset - str_ptr)
        # create new pointer
        new_ptr = cpu_context.mem_alloc(size + 1)
        # Copy at most size bytes of data then add the null terminator.
        cpu_context.mem_copy(str_ptr, new_ptr, size)
        cpu_context.mem_write(new_ptr + size, b"\0")
        return new_ptr


@builtin_func("strlen")
@builtin_func("strnlen_s")
@builtin_func("wcslen")
@builtin_func("wcsnlen_s")
@builtin_func("lstrlenA")
@builtin_func("lstrlenW")
def strlen(cpu_context, func_name, func_args):
    """
    Returns the length of the given null-terminated byte string.
    """
    secure = func_name.endswith("_s")
    wide = func_name.startswith("w") or func_name.endswith("W")
    if secure:
        str_ptr, strsz = func_args
    else:
        (str_ptr,) = func_args
        strsz = None

    if str_ptr:
        logger.debug(" :: Getting length of c string in 0x{:X}".format(str_ptr))
        str = cpu_context.read_data(str_ptr, data_type=WIDE_STRING if wide else STRING)
        size = len(str)
        if wide:
            size /= 2
        # For secure version, strsz is returned if the terminator was not found
        # in the first strsz characters.
        if strsz is not None and size > strsz:
            return strsz
        return size


@builtin_func
def strstr(cpu_context, func_name, func_args):
    """
    Locate substring.

    Returns a pointer to the first occurrence of str2 in str1, or a null pointer if str2 is not part of str1.
    """
    str1_ptr, str2_ptr = func_args
    str1 = cpu_context.read_data(str1_ptr)
    str2 = cpu_context.read_data(str2_ptr)

    offset = str1.find(str2)
    if offset == -1:
        return 0
    else:
        return str1_ptr + offset



"""
DC3-Kordesii framework primary object used for execution of decoders and collection of metadata.
"""

import base64
import codecs
import contextlib
import hashlib
import io
import os
import logging
import shutil
import sys
import tempfile
from typing import List, Optional

import kordesii
from kordesii import decoders, logutil
from kordesii.serialization import deserialize

logger = logging.getLogger(__name__)
ascii_writer = codecs.getwriter("ascii")

# Constant fields
FIELD_DEBUG = "debug"
FIELD_STRINGS = "strings"
FIELD_FILES = "files"


class ReporterLogHandler(logging.Handler):
    """Custom logging handler used to keep backwards compatible with legacy logging mechanism."""

    def __init__(self, reporter):
        super(ReporterLogHandler, self).__init__()
        self._reporter = reporter

    def emit(self, record):
        message = self.format(record)
        if record.levelno > logging.WARNING:
            self._reporter.errors.append(message)
        # Even though reporter uses the name "debug".. This really is an INFO level debug message.
        # (Adding true DEBUG level messages would spam our console.)
        elif logging.INFO <= record.levelno <= logging.WARNING:
            if "debug" not in self._reporter.metadata:
                self._reporter.metadata["debug"] = []
            self._reporter.metadata["debug"].append(message)


class Reporter(object):
    """
    Class for doing decoder execution and metadata reporting.

    This class contains state and data about the current string decoder run, including extracted metadata,
    holding the actual sample, etc.

    Re-using an instance of this class on multiple samples is possible and should be safe, but it is not
    recommended.

    Parameters:
    :param tempdir: sets attribute
    :param disabletempcleanup: disable cleanup (deletion) of temp files

    Attributes:
    :var tempdir: directory where temporary files should be created. Files created in this directory should
        be deleted by decoder. See managed_tempdir for mwcp managed directory
    :var handle: file handle of file to parsed
    :var metadata: Dictionary containing the metadata extracted from the malware by the decoder
    :var errors: list of errors generated by framework. Generally decoders should not set these, they should
        use debug instead
    :var debug: list of debug messages generated by framework or decoders.
    :var strings: list of strings decoded by decoders.
    """

    def __init__(self, tempdir=None, disabletempcleanup=False, base64outputfiles=False):
        self.tempdir = tempdir or tempfile.gettempdir()
        self.metadata = {}
        self.errors = []
        # TODO: Remove disassembler specific details from reporter.
        self.ida_log = ""

        self._log_handler = None
        self._temp_file_name = ""
        self._managed_tempdir = ""

        self._disable_temp_cleanup = disabletempcleanup
        self._base64_output_files = base64outputfiles

        self._other_data = None
        self._deserialized_data = {}

    @property
    def other_data(self):
        """
        The data serialized in the default 'other_data' serializer.

        :return: Dict of the serialized data
        :rtype: dict
        """
        return self.get_serialized("other_data")

    def get_serialized(self, name="other_data"):
        """
        Get the data from a named serializer (that is, other than the default
        'other data' serializer).

        :param str name: The name of the serializer
        :return: Dict of the serialized data
        :rtype: dict
        """
        _deserialized_data = self._deserialized_data
        if name in _deserialized_data:
            return _deserialized_data[name]
        yml_data = self.get_file_contents("{}.yml".format(name))
        data = deserialize(yml_data)
        _deserialized_data[name] = data
        return data

    def managed_tempdir(self):
        """
        Returns the filename of a managed temporary directory. This directory will be deleted when decoder is
        finished, unless tempcleanup is disabled.
        """

        if not self._managed_tempdir:
            self._managed_tempdir = tempfile.mkdtemp(dir=self.tempdir, prefix="kordesii-managed_tempdir-")

            if self._disable_temp_cleanup:
                logger.debug("Using managed temp dir: %s" % self._managed_tempdir)

        return self._managed_tempdir

    def add_string(self, string):
        """
        Record a decoded string
        """
        fieldu = self.convert_to_unicode(FIELD_STRINGS)
        stringu = self.convert_to_unicode(string)

        if fieldu not in self.metadata:
            self.metadata[fieldu] = []

        self.metadata[fieldu].append(stringu)

    def get_strings(self) -> List[str]:
        """
        Get a list of any recorded strings.
        """
        if FIELD_STRINGS in self.metadata:
            return self.metadata[FIELD_STRINGS]

        return []

    def add_output_file(self, filename, data, description=""):
        """
        Add file and its data to metadata.
        """
        fieldu = self.convert_to_unicode(FIELD_FILES)
        filenameu = self.convert_to_unicode(filename)
        descriptionu = self.convert_to_unicode(description)
        md5 = hashlib.md5(data).hexdigest()

        if fieldu not in self.metadata:
            self.metadata[fieldu] = []

        if self._base64_output_files:
            self.metadata[fieldu].append([filenameu, descriptionu, md5, base64.b64encode(data).decode("latin1")])
        else:
            self.metadata[fieldu].append([filenameu, descriptionu, md5])

        if filenameu == u"other_data.yml":
            self.metadata["other_data"] = data.decode("latin1")

    def get_file_contents(self, filename) -> Optional[bytes]:
        """
        If the file name exists and has its contents are stored in the reporter, then take
        the base64 encoded contents, base64 decode it, and return it.
        """
        if FIELD_FILES in self.metadata:
            for entry in self.metadata[FIELD_FILES]:
                if entry[0] == filename and len(entry) == 4:
                    return base64.b64decode(entry[3])

        return None

    def run_decoder(self, name, filename=None, data=None, **run_config):
        """
        Runs specified decoder on file

        :param name: name of decoder module to run
        :param filename: file to parse
        :param data: use data as file instead of loading data from filename
        :param run_config: Extra configuration arguments to pass to kordesii.run_ida()
        """
        self.__reset()

        if not (filename or data):
            raise ValueError("filename or data must be provided.")

        if filename:
            input_file = filename
        else:
            # we were passed data buffer. Lazy initialize a temp file for this
            input_file = os.path.join(self.managed_tempdir(), hashlib.md5(data).hexdigest())
            with open(input_file, "wb") as file_object:
                file_object.write(data)

        try:
            with self.__redirect_stdout():
                found = False
                # TODO: Run all decoders within a single ida call.
                for decoder in kordesii.iter_decoders(name):
                    found = True
                    try:
                        decoder.run(input_file, self, **run_config)
                    except (Exception, SystemExit) as e:
                        logger.exception(
                            "Error running decoder {} on {}".format(decoder.full_name, os.path.basename(input_file))
                        )
                if not found:
                    logger.error("Could not find decoder with name: {}".format(name))
        finally:
            self.__cleanup()

    def convert_to_unicode(self, input_string):
        if isinstance(input_string, str):
            return input_string
        else:
            return str(input_string, encoding="utf8", errors="replace")

    def print_report(self):
        """
        Output in human readable report format
        """
        # Use sys.stdout.buffer if it exists, which is the case for Python 3 and is required
        # for writing a bytes string. Otherwise just write to whatever is at sys.stdout
        print(
            self.get_output_text(),
            # file=ascii_writer(getattr(sys.stdout, 'buffer', sys.stdout), 'backslashreplace')
        )

    def get_output_text(self):
        """
        Get data in human readable report format.
        """

        output = u""

        output += u"----Decoded Strings----\n\n"

        if FIELD_STRINGS not in self.metadata:
            output += u"No decoded strings found\n"
        else:
            for item in self.metadata[FIELD_STRINGS]:
                output += u"{}\n".format(item.encode("unicode-escape").decode())

        if FIELD_FILES in self.metadata:
            output += u"\n----Files----\n\n"
            for item in self.metadata[FIELD_FILES]:
                filename = item[0]
                output += u"{}\n".format(filename)

        if FIELD_DEBUG in self.metadata:
            output += u"\n----Debug----\n\n"
            for item in self.metadata[FIELD_DEBUG]:
                output += u"{}\n".format(item)

        if self.ida_log:
            output += u"\n----IDA Log----\n\n"
            output += u"{}\n".format(self.ida_log)

        if self.errors:
            output += u"\n----Errors----\n\n"
            for item in self.errors:
                output += u"{}\n".format(item)

        return output

    @contextlib.contextmanager
    def __redirect_stdout(self):
        """Redirects stdout temporarily while in a with statement."""
        debug_stdout = io.StringIO()
        orig_stdout = sys.stdout
        sys.stdout = debug_stdout
        try:
            yield
        finally:
            for line in debug_stdout.getvalue().splitlines():
                logger.debug(line)
            sys.stdout = orig_stdout

    def __reset(self):
        """
        Reset all the data in the reporter object that is set during the run_decoder function

        Goal is to make the reporter safe to use for multiple run_decoder instances
        """
        self._temp_file_name = ""
        self._managed_tempdir = ""

        self.metadata = {}
        self.errors = []
        self.ida_log = ""

        # To keep backwards compatibility, setup log handler to add errors and debug messages to reporter.
        # TODO: Remove this when the Reporter object should no longer be responsible for logging.
        log_handler = ReporterLogHandler(self)
        logging.root.addHandler(log_handler)
        # Setup a simple format that doesn't contain any runtime variables.
        log_handler.addFilter(logutil.LevelCharFilter())
        log_handler.setFormatter(logging.Formatter("[%(level_char)s] %(message)s"))
        self._log_handler = log_handler

    def __cleanup(self):
        """
        Cleanup things
        """
        # Remove log handler.
        if self._log_handler:
            logging.root.removeHandler(self._log_handler)
            self._log_handler = None

        # Delete temporary directory.
        if not self._disable_temp_cleanup:
            if self._temp_file_name:
                try:
                    os.remove(self._temp_file_name)
                except Exception as e:
                    logger.warning("Failed to purge temp file: %s, %s" % (self._temp_file_name, str(e)))
                self._temp_file_name = ""

            if self._managed_tempdir:
                try:
                    shutil.rmtree(self._managed_tempdir, ignore_errors=True)
                except Exception as e:
                    logger.warning("Failed to purge temp dir: %s, %s" % (self._managed_tempdir, str(e)))
                self._managed_tempdir = ""

        self._temp_file_name = ""
        self._managed_tempdir = ""

    def __del__(self):
        self.__cleanup()



"""
Extra utility functions for interfacing with ida.
"""

# TODO: For lack of a better name, this is called "utils". Change this.

from __future__ import absolute_import

from typing import Union, List

import re
import logging

import idaapi
import ida_bytes
import ida_entry
import ida_funcs
import ida_nalt
import idc
import idautils


logger = logging.getLogger(__name__)


READ_LENGTH = 65536


def iter_imports(module_name=None, api_names=None):
    """
    Iterate the thunk function wrappers for API imports.
    Yields the module name, function name, and reference to function.

    .. code_block:: python

        for ea, name, module_name in utils.iter_imports():
            print("{}.{} function at: 0x{:0x}".format(module_name, name, ea))

        for ea, name, _ in utils.iter_imports("KERNEL32"):
            print("KERNEL32.{} function at: {}".format(name, ea))

        for ea, name, module_name in utils.iter_imports(api_names=["GetProcAddress", "GetFileSize"]):
            print("{}.{} function at: {}".format(module_name, name, ea))

        for ea, _, _ in utils.iter_imports("KERNEL32", "GetProcAddress"):
            print("KERNEL32.GetProcAddress function at: {}".format(ea))

    NOTE: The same function name can be yield more than once if it
    appears in multiple modules or has multiple thunk wrappers.

    Name is the original import name and does not necessarily reflect the function name.
    e.g. "GetProcAddress", "GetProcAddress_0", and "GetProcAddress_1" will all be "GetProcAddress"

    :param module_name: Filter imports to a specified library.
    :param api_names: Filter imports to specific API name(s).
        Can be a string of a single name or list of names.

    :yield: (ea, api_name, module_name)
    """
    if isinstance(api_names, str):
        api_names = [api_names]

    for i in range(ida_nalt.get_import_module_qty()):
        _module_name = ida_nalt.get_import_module_name(i)
        if not _module_name:
            continue
        if module_name and module_name.lower() != _module_name.lower():
            continue

        entries = []
        target_set = set(api_names) if api_names else None

        def callback(ea, name, ordinal):
            if name:
                # Sometimes IDA includes "__imp_" to the front of the name.
                # Strip this off to be more consistent to what you would see in the GUI.
                if name.startswith("__imp_"):
                    name = name[6:]

                # Collect name if matches filter or if no filter set.
                if target_set and (
                    name in target_set
                    or name.strip("_") in target_set
                    or any(re.match("_*{}_+[0-9]?".format(name_), name) for name_ in target_set)
                ):
                    entries.append((ea, name))
                    target_set.difference_update({name, name.strip("_")})
                    if not target_set:
                        # Found all targeted function names. stop enumeration.
                        return False
                elif not api_names:
                    entries.append((ea, name))
            return True  # continue enumeration

        ida_nalt.enum_import_names(i, callback)

        for ea, name in entries:
            # Yield thunk wrapper functions if they exists.
            for xref in idautils.XrefsTo(ea):
                func = ida_funcs.get_func(xref.frm)
                if not func:
                    continue
                if func.flags & ida_funcs.FUNC_THUNK:
                    yield xref.frm, name, _module_name

            # Yield reference in data segment signature
            # (yielding after thunks, since those are more likely to be used)
            yield ea, name, _module_name


def iter_exports():
    """
    Iterate API exports.

    :yield: (ea, name)
    """
    for i in range(ida_entry.get_entry_qty()):
        ordinal = ida_entry.get_entry_ordinal(i)
        ea = ida_entry.get_entry(ordinal)
        name = ida_entry.get_entry_name(ordinal)
        yield ea, name


def iter_functions(func_names: Union[None, str, List[str]] = None):
    """
    Iterate all defined functions and yield their address and name.
    (This includes imported functions)

    :param func_names: Filter based on specific function names.

    :yield: (ea, name)
    """
    if isinstance(func_names, str):
        func_names = [func_names]

    # Yield declared functions.
    for ea in idautils.Functions():
        name = idc.get_func_name(ea)
        if (
            not func_names
            or name in func_names
            or name.strip("_") in func_names
            or any(re.match("_*{}_[0-9]?".format(name_), name) for name_ in func_names)
        ):
            yield ea, name

    # Also yield from imported.
    for ea, name, _ in iter_imports(api_names=func_names):
        yield ea, name


def get_import_addr(api_name, module_name=None):
    """
    Returns the first instance of a function that wraps the given API name.

    .. code_block:: python

        proc_func_ea = get_import_addr("GetProcAddress")

    :param api_name: Name of API
    :param module_name: Library of API

    :returns: Address of function start or None if not found.
    """
    for ea, _, _ in iter_imports(module_name, api_name):
        return ea


def get_export_addr(export_name):
    """
    Return the location of an export by name

    :param export_name: Target export

    :return: Location of target export or None
    """
    for ea, name in iter_exports():
        if name == export_name:
            return ea


def get_function_addr(func_name: str):
    """
    Obtain a function in the list of functions for the application by name.
    Supports using API resolved names if necessary.

    :param func_name: Name of function to obtain

    :return: start_ea of function or None
    """
    for ea, _ in iter_functions(func_name):
        return ea


def lines(start=None, end=None, reverse=False, max_steps=None):
    """
    Iterates through instructions within the start address and end addresses.

    :param start: Address of the starting instruction. (starts at beginning if not defined)
    :param end: Address of the end instruction.
    :param reverse: Iterates up if true.
    :param max_steps: Maximum number of steps to iterate.
    :yields: instructions addresses
    """
    max_ea = idaapi.cvar.inf.maxEA
    min_ea = idaapi.cvar.inf.minEA

    # Normalize start and end addresses.
    if reverse:
        if start is None:
            start = max_ea - 1
        if end is None:
            end = 0
        start = max(start, end)
    else:
        if start is None:
            start = min_ea
        if end is None:
            end = max_ea + 1
        start = min(start, end)

    steps = 0
    func = ida_bytes.prev_head if reverse else ida_bytes.next_head
    ea = ida_bytes.get_item_head(start)
    while ea != idc.BADADDR:
        yield ea
        ea = func(ea, end)
        steps += 1
        if max_steps and steps >= max_steps:
            break


def get_string(ea: int) -> bytes:
    """
    Returns a string from the given location.

    :param ea: starting address of string

    :return: A string
    """
    stype = idc.get_str_type(ea)
    return idc.get_strlit_contents(ea, strtype=stype)



"""
Interface for function management.
"""

import logging
import re

import ida_idp
import ida_nalt
import ida_typeinf
import idc

from . import utils

logger = logging.getLogger(__name__)


class FunctionSignature(object):
    """
    Interface for a function signature.

    NOTE: This object retrieves the signature on initialization.
    Any external changes to the function's signature after this object is created will take
    no affect.
    As well, any changes done to this object will not affect the signature in the IDB unless
    the apply() function is called.
    """

    def __init__(self, cpu_context, start_ea):
        """
        :param cpu_context: ProcessorContext to use for pulling argument values.
        :param start_ea: Starting address of function to create function signature from.

        :raises RuntimeError: If a function type could not be created from given ea.
        """
        self._cpu_context = cpu_context
        self.start_ea = start_ea
        # TODO: Possibly move the get_function_data work into this class?
        self._func_type_data = utils.get_function_data(self.start_ea)
        tif = ida_typeinf.tinfo_t()
        ida_nalt.get_tinfo(tif, self.start_ea)
        self._tif = tif

    def __repr__(self):
        return "< FunctionSignature : {} >".format(self.declaration)

    @property
    def name(self):
        """The demangled name of function."""
        # TODO: Move the get_function_name() into this class?
        return utils.get_function_name(self.start_ea)

    @property
    def declaration(self):
        """The full function declaration."""
        # There is no way to get the full function declaration directly from IDA with
        # the function name intact. So we have to recreate it.

        # If function doesn't have a name (usually because the function was dynamically created
        # within a register), then we are just going to call it "no_name" so we can still get the
        # function typing to still work.

        return re.sub("\(", " {}(".format(self.name or "no_name"), "{};".format(str(self._tif)))

    @declaration.setter
    def declaration(self, decl):
        """
        Changes the declaration of the function internally.
        """
        # Ensure ends with ';'
        if not decl.endswith(";"):
            decl += ";"

        tif = ida_typeinf.tinfo_t()
        til = ida_typeinf.get_idati()
        func_type_data = ida_typeinf.func_type_data_t()
        ida_typeinf.parse_decl(tif, til, decl, ida_typeinf.PT_SIL)
        tif.get_func_details(func_type_data)
        self._tif = tif
        self._func_type_data = func_type_data

    def apply(self):
        """Applies the currently set signature to the IDB."""
        idc.apply_type(self.start_ea, self.declaration)

    @property
    def arg_types(self):
        """Tuple of the argument types."""
        return tuple(arg.declaration for arg in self.args)

    @arg_types.setter
    def arg_types(self, arg_types):
        """Set tuple of argument types."""
        self.declaration = re.sub("\(.*\)", "({})".format(",".join(arg_types)), self.declaration)

    @property
    def args(self):
        return [FunctionArg(self._cpu_context, i, funcarg_obj) for i, funcarg_obj in enumerate(self._func_type_data)]


class FunctionArg(object):
    """
    Interface for a function argument from FunctionSignature
    """

    # TODO: Figure out how to do this without using funcarg_t object.
    def __init__(self, cpu_context, idx, funcarg_obj):
        self._cpu_context = cpu_context
        self._funcarg_obj = funcarg_obj
        self.idx = idx

    def __repr__(self):
        return "< FunctionArg : {} = {!r} >".format(self.declaration, self.value)

    @property
    def width(self):
        return self._funcarg_obj.type.get_size()

    @property
    def name(self):
        return self._funcarg_obj.name

    @name.setter
    def name(self, value):
        self._funcarg_obj.name = value

    @property
    def type(self):
        """User friendly type name."""
        return str(self._funcarg_obj.type)

    @type.setter
    def type(self, value):
        """
        Sets FunctionArg to a new type.

        NOTE: Setting the type here has no affect on the FunctionSignature object this came from.
        """
        is_ptr = value.endswith("*")
        value = value.strip(" *")

        # Create new tinfo object of type.
        tif = ida_typeinf.tinfo_t()
        tif.get_named_type(ida_typeinf.get_idati(), value)

        # If a pointer, create another tinfo object that is the pointer of the first.
        if is_ptr:
            tif2 = ida_typeinf.tinfo_t()
            tif2.create_ptr(tif)
            tif = tif2

        self._funcarg_obj.type = tif

    @property
    def declaration(self):
        """Argument type declaration."""
        return " ".join([self.type, self.name])

    @property
    def value(self):
        # TODO: Pull value data based on type.

        argloc = self._funcarg_obj.argloc
        loc_type = argloc.atype()

        # This type occurs when we have created an uninitialized argument.
        if loc_type == ida_typeinf.ALOC_NONE:
            logger.warning("Argument {} location is of type ALOC_NONE".format(self.idx))
            return None

        elif loc_type == ida_typeinf.ALOC_STACK:
            # read the argument from the stack using the calculated stack offset from the disassembler
            cur_esp = self._cpu_context.sp + argloc.stkoff()
            value = self._cpu_context.memory.read(cur_esp, self._cpu_context.byteness)
            return utils.struct_unpack(value)

        elif loc_type == ida_typeinf.ALOC_DIST:  # arguments described by multiple locations
            # TODO: Uses the scattered_aloc_t class, which is a qvector or argpart_t objects
            # argloc.scattered()
            raise NotImplementedError("Argument {} location of type ALOC_DIST".format(self.idx))

        elif loc_type == ida_typeinf.ALOC_REG1:  #  single register
            # TODO: Determine better way to convert reg1 integer to name.
            reg_name = ida_idp.get_reg_name(argloc.reg1(), self.width)
            return self._cpu_context.registers[reg_name]

        elif loc_type == ida_typeinf.ALOC_REG2:  # register pair (eg: edx:eax [reg2:reg1])
            # TODO: CURRENTLY UNTESTED
            logger.info("Argument {} of untested type ALOC_REG2.  " "Verify results and report issues".format(self.idx))
            # TODO: Assuming registers are the same width..
            #  need to determine if that is an accurate assumption.
            reg1_name = ida_idp.get_reg_name(argloc.reg1(), self.width)
            reg2_name = ida_idp.get_reg_name(argloc.reg2(), self.width)
            reg1_value = self._cpu_context.registers[reg1_name]
            reg2_value = self._cpu_context.registers[reg2_name]
            return reg2_value << self.width | reg1_value

        elif loc_type == ida_typeinf.ALOC_RREL:  # register relative (displacement from address pointed by register
            # TODO: CURRENTLY UNTESTED
            logger.info(
                "Argument {} of untested type ALOC_RREL.  " "Verify results and report issues.".format(self.idx)
            )
            # Obtain the register-relative argument location
            rrel = argloc.get_rrel()
            reg_name = ida_idp.get_reg_name(rrel.reg, self.width)
            value = self._cpu_context.registers[reg_name]
            return value + rrel.offset

        elif loc_type == ida_typeinf.ALOC_STATIC:  # global address
            # TODO: CURRENTLY UNTESTED
            logger.info(
                "Argument {} of untested type ALOC_STATIC.  " "Verify results and report issues.".format(self.idx)
            )
            return argloc.get_ea()

        elif loc_type >= ida_typeinf.ALOC_CUSTOM:  #  custom argloc
            # TODO: Will need to figure out the functionality and usage for the custloc_desc_t structure
            # argloc.get_custom()
            raise NotImplementedError("Argument {} location of type ALOC_CUSTOM".format(self.idx))



"""
DC3-Kordesii framework command line tool.

Used for running and testing decoders.
"""

import datetime
import glob
import hashlib
import json
import logging
import os
import shutil
import sys
import timeit
import traceback

import click
import kordesii
import tabulate
from kordesii.tester import Tester

logger = logging.getLogger("kordesii")


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.option("-d", "--debug", is_flag=True, help="Enables DEBUG level logs.")
@click.option("-v", "--verbose", is_flag=True, help="Enables INFO level logs.")
@click.option(
    "--decoder-dir",
    type=click.Path(exists=True, file_okay=False),
    help="Optional extra decoder directory.",
    envvar="KORDESII_DECODER_DIR",
    show_envvar=True,
)
@click.option(
    "--decoder-source",
    help="Set a default decoder source to use. " "If not, registered decoders from all sources will be available.",
    envvar="KORDESII_DECODER_SOURCE",
    show_envvar=True,
)
def main(debug, verbose, decoder_dir, decoder_source):
    # Setup logging
    kordesii.setup_logging()
    if debug:
        logging.root.setLevel(logging.DEBUG)
    elif verbose:
        logging.root.setLevel(logging.INFO)
    # else let log_config.yaml set log level.

    # Register parsers
    kordesii.register_entry_points()
    if decoder_dir:
        kordesii.register_decoder_directory(decoder_dir)

    if decoder_source:
        kordesii.set_default_source(decoder_source)


@main.command("list")
@click.option("-j", "--json", "json_", is_flag=True, help="Display as JSON output.")
def list_(json_):
    """Lists registered decoders."""
    descriptions = kordesii.get_decoder_descriptions()
    if json_:
        print(json.dumps(descriptions, indent=4))
    else:
        print(tabulate.tabulate(descriptions, headers=["NAME", "SOURCE", "AUTHOR", "DESCRIPTION"]))


@main.command()
@click.argument("decoder", required=True)
@click.argument("input", nargs=-1, type=click.Path())
@click.option("-j", "--json", "json_", is_flag=True, help="Display as JSON output.")
# TODO: We can't allow user to change decoder output directory until a bit more refactoring is done.
# @click.option('-o', '--output-dir', type=click.Path(exists=True, file_okay=False),
#               help='Output directory.')
@click.option(
    "--output-files/--no-output-files", default=True, show_default=True, help="Whether to output files to filesystem."
)
@click.option(
    "--cleanup/--no-cleanup", default=True, show_default=True, help="Whether to cleanup temporary files after parsing."
)
@click.option(
    "-m",
    "--tempdir",
    type=click.Path(file_okay=False),
    help="Custom directory to store temporary files. Useful if you combine this with --no-cleanup.",
)
# Disassembler specific options:
@click.option("-e", "--enable-ida-log", is_flag=True, help="Include the log contents produced by IDA in the results.")
@click.option(
    "-t",
    "--timeout",
    type=int,
    default=0,
    show_default=True,
    help="Timeout for running IDA. A timeout of 0 disables the timeout.",
)
def parse(decoder, input, json_, output_files, cleanup, tempdir, enable_ida_log, timeout):
    """
    Parses given input with given parser.

    \b
    DECODER: Name of decoder to run.
    INPUT: One or more input file paths. (Wildcards are allowed).

    \b
    Common usages::
        kordesii parse foo ./malware.bin                         - Run foo decoder on ./malware.bin
        kordesii parse foo ./repo/*                              - Run foo decoder on files found in repo directory.
        kordesii parse --json foo ./malware.bin                  - Run foo decoder and display results as json.
    """
    # Python won't process wildcards when used through Windows command prompt.
    if any("*" in path for path in input):
        new_input = []
        for path in input:
            if "*" in path:
                new_input.extend(glob.glob(path))
            else:
                new_input.append(path)
        input = new_input

    input_files = list(filter(os.path.isfile, input))
    if not input_files:
        sys.exit("Unable to find any input files.")

    # Run Kordesii
    try:
        reporter = kordesii.Reporter(tempdir=tempdir, disabletempcleanup=not cleanup,)
        results = []
        for path in input_files:
            logger.info("Parsing: {}".format(path))
            input_file = os.path.abspath(path)
            reporter.run_decoder(
                decoder,
                input_file,
                timeout=timeout,
                log=enable_ida_log,
                cleanup_txt_files=cleanup,
                cleanup_idb_files=cleanup,
                cleanup_output_files=not output_files,
            )
            # TODO: Pull errors and ida logs from logger?
            result = reporter.metadata
            if reporter.errors:
                result["errors"] = reporter.errors
            if reporter.ida_log:
                result["ida_log"] = reporter.ida_log
            results.append(result)
            if not json_:
                reporter.print_report()

        if json_:
            print(json.dumps(results, indent=4))

    except Exception as e:
        error_message = "Error running DC3-Kordesii: {}".format(e)
        traceback.print_exc()
        if format == "json":
            print(json.dumps({"errors": [error_message]}))
        else:
            print(error_message)
        sys.exit(1)


def _median(data):
    """
    'Borrowed' from Py3's statistics library.

    :param data: Data to get median of
    :return: Median as a float or int
    :rtype: float or int
    """
    data = sorted(data)
    length = len(data)
    if length == 0:
        raise ValueError("No median for empty data.")
    elif length % 2 == 1:
        return data[length // 2]
    else:
        i = length // 2
        return (data[i - 1] + data[i]) / 2


def _run_tests(tester, silent=False, show_passed=False):
    print("Running test cases. May take a while...")

    start_time = timeit.default_timer()
    test_results = []
    total = tester.total
    failed = []

    # Generate format string.
    digits = len(str(total))
    if not tester.test_cases:
        decoder_len = 10
        filename_len = 10
    else:
        decoder_len = max(len(test_case.decoder_name) for test_case in tester.test_cases)
        filename_len = max(len(test_case.filename) for test_case in tester.test_cases)
    msg_format = "{{decoder:{0}}} {{filename:{1}}} {{run_time:.4f}}s".format(decoder_len, filename_len)

    format_str = "{{count:> {0}d}}/{{total:0{0}d}} - ".format(digits) + msg_format

    # Run tests and output progress results.
    for count, test_result in enumerate(tester, start=1):
        if not test_result.passed:
            failed.append((count, test_result.decoder_name, test_result.filename))

        if test_result.run_time:  # Ignore missing tests from stat summary.
            test_results.append(test_result)

        if not silent:
            message = format_str.format(
                count=count,
                total=total,
                decoder=test_result.decoder_name,
                filename=test_result.filename,
                run_time=test_result.run_time,
            )
            # Skip print() to immediately flush stdout buffer (issue in Docker containers)
            sys.stdout.write(message + "\n")
            sys.stdout.flush()
            if not test_result.passed or show_passed:
                test_result.print()

    end_time = timeit.default_timer()

    # Present test statistics
    if not silent and test_results:
        print("\nTest stats:")
        print("\nTop 10 Slowest Test Cases:")

        format_str = "{index:2}. " + msg_format

        # Cases sorted slowest first
        sorted_cases = sorted(test_results, key=lambda x: x.run_time, reverse=True)
        for i, test_result in enumerate(sorted_cases[:10], start=1):
            print(
                format_str.format(
                    index=i,
                    decoder=test_result.decoder_name,
                    filename=test_result.filename,
                    run_time=test_result.run_time,
                )
            )

        print("\nTop 10 Fastest Test Cases:")
        for i, test_result in enumerate(list(reversed(sorted_cases))[:10], start=1):
            print(
                format_str.format(
                    index=i,
                    decoder=test_result.decoder_name,
                    filename=test_result.filename,
                    run_time=test_result.run_time,
                )
            )

        run_times = [test_result.run_time for test_result in test_results]
        print("\nMean Running Time: {:.4f}s".format(sum(run_times) / len(test_results)))
        print("Median Running Time: {:.4f}s".format(_median(run_times)))
        print("Cumulative Running Time: {}".format(datetime.timedelta(seconds=sum(run_times))))
        print()

    print("Total Running Time: {}".format(datetime.timedelta(seconds=end_time - start_time)))

    if failed:
        print()
        print("Failed tests:")
        for test_info in failed:
            print("#{} - {}\t{}".format(*test_info))
        print()

    print("All Passed = {0}\n".format(not failed))
    exit(0 if not failed else 1)


def _get_malware_repo_path(file_path, malware_repo):
    """
    Gets file path for a file in the malware_repo based on the md5 of the given file_path.
    """
    with open(file_path, "rb") as fo:
        md5 = hashlib.md5(fo.read()).hexdigest()
    return os.path.join(malware_repo, md5[0:4], md5)


def _add_to_malware_repo(file_path, malware_repo):
    """
    Adds the given file path to the malware repo.
    Returns resulting destination path.
    """
    dest_path = _get_malware_repo_path(file_path, malware_repo)
    dest_dir = os.path.dirname(dest_path)

    if os.path.isfile(dest_path):
        click.echo("File already exists: {}".format(dest_path))
        return dest_path

    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)

    click.echo("Copying {} to {}".format(file_path, dest_path))
    shutil.copy(file_path, dest_path)
    return dest_path


@main.command()
@click.option(
    "-t",
    "--testcase-dir",
    type=click.Path(exists=True, file_okay=False),
    help="Directory containing JSON test case files. (defaults to a "
    '"tests" directory located within the root of the decoders directory)',
    envvar="KORDESII_TESTCASE_DIR",
    show_envvar=True,
)
@click.option(
    "-m",
    "--malware-repo",
    type=click.Path(exists=True, file_okay=False),
    help="Directory containing malware samples used for testing.",
    envvar="KORDESII_MALWARE_REPO",
    show_envvar=True,
)
# Arguments used for run test cases.
@click.option(
    "-n", "--nprocs", type=int, help="Number of test cases to run simultaneously. [default: 3/4 * logical CPU cores]"
)
# Arguments used to generate and update test cases
@click.option("-u", "--update", is_flag=True, help="Update all stored test cases with newly produced results.")
@click.option(
    "-a",
    "--add",
    multiple=True,
    type=click.Path(exists=True, dir_okay=False),
    help="Adds given file to the test case. (Will first copy file to malware repo if provided.)",
)
@click.option(
    "-i",
    "--add-filelist",
    multiple=True,
    type=click.Path(exists=True, dir_okay=False),
    help="Adds a file of file paths separated by newlines to the test case.",
)
@click.option(
    "-x",
    "--delete",
    multiple=True,
    type=click.Path(exists=True, dir_okay=False),
    help="Deletes given file from the test case. "
    "(Note, this does not delete the test file if placed in a malware repo.)",
)
@click.option("-y", "--yes", is_flag=True, help="Auto confirm questions.")
@click.option("--force", is_flag=True, help="Force test case add/update when errors are encountered.")
# Arguments to configure console output
@click.option(
    "-f",
    "--show-passed",
    is_flag=True,
    help="Display tests case details for passed tests as well." "By default only failed tests are shown.",
)
@click.option("-s", "--silent", is_flag=True, help="Limit output to statement saying whether all tests passed or not.")
# Decoder to process.
@click.argument("decoder", nargs=-1, required=False)
def test(
    testcase_dir, malware_repo, nprocs, update, add, add_filelist, delete, yes, force, show_passed, silent, decoder
):
    """
    Testing utility to create and execute decoder test cases.

    \b
    DECODER: Decoders to test. Test all decoders if not provided.

    \b
    Common usages::
        kordesii test                                             - Run all tests cases.
        kordesii test foo                                         - Run test cases for foo decoder.
        kordesii test foo -u                                      - Update existing test cases for foo decoder.
        kordesii test -u                                          - Update existing test cases for all decoders.
        kordesii test foo --add=./malware.bin                     - Add test case for malware.bin sample for foo parser.
        kordesii test foo --add-filelist=./paths.txt              - Add tests cases for foo decoder using text file of paths.
        kordesii test foo --delete=./malware.bin                  - Delete test case for malware.bin sample for foo parser.
    """
    # Configure test object
    reporter = kordesii.Reporter()
    tester = Tester(reporter, results_dir=testcase_dir, decoder_names=decoder or [None], nprocs=nprocs,)

    # Add/Delete
    if add or add_filelist or delete:
        if not decoder:
            # Don't allow adding a file to ALL test cases.
            raise click.BadParameter("DECODER must be provided when adding or deleting a file from a test case.")

        # Cast tuple to list so we can manipulate.
        add = list(add)
        for filelist in add_filelist:
            with open(filelist, "r") as f:
                for file_path in f.readlines():
                    add.append(file_path.rstrip("\n"))

        for file_path in add:
            click.echo("Adding new test cases. May take a while...")
            if malware_repo:
                file_path = _add_to_malware_repo(file_path, malware_repo)
            tester.add_test(file_path, force)

        for file_path in delete:
            if malware_repo:
                file_path = _get_malware_repo_path(file_path, malware_repo)
            tester.remove_test(file_path)

    # Update
    elif update:
        if not decoder and not yes:
            click.confirm("WARNING: About to update test cases for ALL decoders. Continue?", abort=True)
        click.echo("Updating test cases. May take a while...")
        tester.update_tests(force)

    # Run tests
    else:
        if not decoder and not yes:
            click.confirm("DECODER argument not provided. Run tests for ALL decoders?", default=True, abort=True)
        # Force ERROR level logs so we don't spam the console.
        logging.root.setLevel(logging.ERROR)
        _run_tests(tester, silent, show_passed)


@main.command()
@click.option("--host", default="127.0.0.1", show_default=True, help="The interface to bind to.")
@click.option("--port", default=8081, show_default=True, help="The port to bind to.")
@click.option(
    "--debug", is_flag=True, help="Show the interactive debugger if errors occur, and auto-reload on code changes."
)
def serve(host, port, debug):
    """Run a server to handle parsing requests."""
    from kordesii.tools import server

    if debug:
        os.environ["FLASK_ENV"] = "development"

    app = server.create_app()
    app.run(host=host, port=port, use_reloader=False)


if __name__ == "__main__":
    main(sys.argv[1:])



"""
Provides user-facing API for function_tracing.
"""

# Import functions and classes we want to expose as the API.
from .function_tracer import FunctionTracer, TracerCache, get_tracer, hook_tracers, clear_hooks
from .flowchart import FlowChart, CustomBasicBlock
from .exceptions import *

# Expose constants that may be useful
from .constants import *

# TODO: Can these be generalized for all architectures?
from .x86_64 import STACK_BASE, RSP_OFFSET, RBP_OFFSET

# Import architecture packages to ensure they are registered.
from . import x86_64, ARM



"""
This module uses the idaapi.FlowChart object and extends it as well as idaapi.BasicBlock in order to add
functionality including Breadth-First and Depth-First chart traversal, locating a specific block within the
chart based on an EA, generating a list of all possible paths to a specified EA, etc.
"""

import functools
import logging
from operator import attrgetter
from copy import copy, deepcopy

import idaapi
import idautils
import idc

from .cpu_context import ProcessorContext


logger = logging.getLogger(__name__)


class PathNode(object):
    """
    Represents a linked-list of objects constituting a path from a specific node to the function entry point node.  This
    object can also track cpu context up to a certain EA.
    """

    _cache = {}

    def __init__(self, bb, prev):
        self.bb = bb
        self.prev = prev
        # TODO: Add caching for multiple init_contexts?
        self._context = None
        self._context_ea = None  # ea that the context has been filled to (but not including)
        self._init_context = None  # the context used at the starting path

    @classmethod
    def from_cache(cls, bb, prev):
        """Constructor that caches and reuses existing instances."""
        try:
            return cls._cache[(bb, prev)]
        except KeyError:
            path_node = cls(bb, prev)
            cls._cache[(bb, prev)] = path_node
            return path_node

    def __contains__(self, ea):
        return ea in self.bb

    def __repr__(self):
        return "PathNode({!r})".format(self.bb)

    def path(self):
        """Returns a list of PathNode objects represented by the linked list."""
        if self.prev:
            return self.prev.path() + [self]
        else:
            return [self]

    def cpu_context(self, ea=None, init_context=None):
        """
        Returns the cpu context filled to (but not including) the specified ea.

        :param int ea: address of interest (defaults to the last ea of the block)
        :param init_context: Initial context to use for the start of the path.
            (defaults to an new empty context)

        :return cpu_context.ProcessorContext: cpu context
        """
        if ea is not None and not (self.bb.start_ea <= ea < self.bb.end_ea):
            raise KeyError(
                "Provided address 0x{:X} not in this block "
                "(0x{:X} :: 0x{:X})".format(ea, self.bb.start_ea, self.bb.end_ea)
            )

        # Determine address to stop computing.
        if ea is None:
            end = self.bb.end_ea  # end_ea of a BasicBlock is the first address after the last instruction.
        else:
            end = ea

        # Determine if we need to force the creation of a new context if we have a different init_context.
        new_init_context = self._init_context != init_context
        self._init_context = init_context

        assert end is not None
        # Fill context up to requested endpoint.
        if self._context_ea != end or new_init_context:
            # Create context if:
            #   - not created
            #   - current context goes past requested ea
            #   - given init_context is different from the previously given init_context.
            if not self._context or self._context_ea > end or new_init_context:
                # Need to check if there is a prev, if not, then we need to create a default context here...
                if self.prev:
                    self._context = self.prev.cpu_context(init_context=init_context)
                    # Modify the context for the current branch if required
                    self._context.prep_for_branch(self.bb.start_ea)
                elif init_context:
                    self._context = deepcopy(init_context)
                else:
                    self._context = ProcessorContext.from_arch()

                self._context_ea = self.bb.start_ea

            if self._context_ea != end:
                # Fill context up to requested ea.
                logger.debug("Emulating instructions 0x{:08X} -> 0x{:08X}".format(self._context_ea, end))
                for ip in idautils.Heads(self._context_ea, end):
                    self._context.execute(ip)

            self._context_ea = end

        # Set the next instruction pointer to be the end instruction that we did NOT execute.
        self._context.ip = end

        return deepcopy(self._context)


def _get_flowchart(ea):
    """
    Helper function to obtain an idaapi.FlowChart object for a given ea.

    :param int ea: ea of interest

    :return idaapi.FlowChart: idaapi.FlowChart object
    """
    func = idaapi.get_func(ea)
    flowchart_ = idaapi.FlowChart(func)
    return flowchart_


def _get_codeblock(ea):
    """
    Helper function to obtain a idaapi.BasicBlock object containing a given ea.

    :param int ea: ea of interest

    :return idaapi.BasicBlock: idaapi.BasicBlock object
    """
    flowchart_ = _get_flowchart(ea)
    for code_block in flowchart_:
        if code_block.start_ea <= ea < code_block.end_ea:
            return code_block


@functools.total_ordering
class CustomBasicBlock(idaapi.BasicBlock):
    """
    An idaapi.BasicBlock object which has been extended with additional functionality beyond the base class.

    Additional functionality:
        - Ability to use BasicBlocks as hashable objects (ie: as dictionary keys)
        - Check if two BasicBlocks are equal (based on their start_ea)
        - Check if an EA is contained in a BasicBlock (ie: if ea in <CustomBasicBlock>:)
        - Check length of block
        - Iterator addresses in block.
        - Iterator of paths leading to this block.
    """

    def __init__(self, id_or_ea, bb=None, fc=None):
        if bb is None and fc is None:
            temp_codeblock = _get_codeblock(id_or_ea)
            self.__dict__.update(temp_codeblock.__dict__)
        else:
            super(CustomBasicBlock, self).__init__(id=id_or_ea, bb=bb, fc=fc)

    def __hash__(self):
        return self.start_ea

    def __repr__(self):
        return "<CustomBasicBlock(start_ea=0x{:08X}, end_ea=0x{:08X})>".format(self.start_ea, self.end_ea)

    def __eq__(self, other):
        return self.start_ea == other.start_ea

    def __lt__(self, other):
        return self.start_ea < other.start_ea

    def __contains__(self, ea):
        return self.start_ea <= ea < self.end_ea

    def __len__(self):
        """Length of block is the number of instructions contained within."""
        return len(list(self.heads()))

    def heads(self, start=None, reverse=False):
        """
        Iterates all the heads within the given block.

        :param start: Start address (defaults to start_ea or end_ea)
        :param reverse: Direction to iterate

        :yields: Instruction addresses.

        :raises ValueError: If given start address it not in block.
        """
        if start and start not in self:
            raise ValueError("Start address 0x{:08X} is not in block: {!r}".format(start, self))

        if reverse:
            for head in reversed(list(idautils.Heads(self.start_ea, start or self.end_ea))):
                yield head
        else:
            for head in idautils.Heads(start or self.start_ea, self.end_ea):
                yield head

    def paths(self, _visited=None):
        """
        Iterates the paths that lead to this block.

        :param _visited: Internally used.
        :yields: PathNode objects that represent the last entry of the path linked list.
        """
        if _visited is None:
            _visited = set()

        # Otherwise generate path nodes and cache results for next time.
        _visited.add(self.start_ea)

        parents = list(self.preds())
        if not parents:
            yield PathNode.from_cache(self, prev=None)
        else:
            for parent in parents:
                if parent.start_ea in _visited:
                    continue

                # Create path nodes for each path of parent.
                for parent_path in parent.paths(_visited=_visited):
                    yield PathNode.from_cache(self, prev=parent_path)

        _visited.remove(self.start_ea)


class FlowChart(idaapi.FlowChart):
    """
    Object containing the function graph generated by IDA.  Implements the traversal of the function.
    """

    def __init__(self, f, bounds=None, flags=idaapi.FC_PREDS):
        self.f = idaapi.get_func(f)
        super(FlowChart, self).__init__(f=self.f, bounds=bounds, flags=flags)

    def _traverse(self, start_ea=None, dfs=False):
        """
        Blind traversal of the graph.
        For each block, obtain the children (or blocks which are reachable
        from the current block), sort the children by their start_ea in ascending order, and "push" the list on to the
        front of the non_visisted blocks list.

        :param int start_ea: EA within a block from which to start traversing
        :param bool dfs: If true, traversal will be depth-first. If false, traversal will be breadth-first.

        :yield CustomBasicBlock: function block object
        """
        # Set our flag to True if start_ea is none so we yield all blocks, else wait till we find the requested block
        block_found = start_ea is None
        non_visited = [self[0]]
        visited = set()
        while non_visited:
            cur_block = non_visited.pop(0)
            if hash(cur_block) in visited:
                continue

            visited.add(hash(cur_block))
            succs = sorted(cur_block.succs())
            if dfs:
                # [0:0] allows us to extend to the front
                non_visited[0:0] = succs
            else:
                non_visited.extend(succs)

            if not block_found:
                block_found = start_ea in cur_block

            if block_found:
                yield cur_block

    def _traverse_reverse(self, start_ea=None, dfs=False):
        """
        Perform a reverse traversal of the graph in depth-first/breadth-first manner where given a start node, traverse 1 complete
        path to the root node before following additional paths.

        :param int start_ea: EA within a block from which to start traversing
        :param bool dfs: If true, traversal will be depth-first. If false, traversal will be breadth-first.

        :yield: function block object
        """
        if start_ea:
            non_visited = [self.find_block(start_ea)]
        else:
            non_visited = list(sorted(self, key=attrgetter("start_ea")))[-1:]

        visited = set()
        while non_visited:
            cur_block = non_visited.pop(0)
            if hash(cur_block) in visited:
                continue

            visited.add(hash(cur_block))

            preds = sorted(cur_block.preds(), reverse=True)
            # For now, only consider predicates that are before the current block.
            # This helps to prevent cyclic loops.
            preds = [pred for pred in preds if pred < cur_block]
            if dfs:
                non_visited[0:0] = preds
            else:
                non_visited.extend(preds)

            yield cur_block

    def blocks(self, start=None, reverse=False, dfs=False):
        """
        Iterates over CustomBasicBlocks.

        :param int start: optional address to start iterating from.
        :param bool reverse: iterate in reverse
        :param bool dfs: If true, traversal will be depth-first. If false, traversal will be breadth-first.
        """
        if reverse:
            for cur_block in self._traverse_reverse(start, dfs=dfs):
                yield cur_block

        else:
            for cur_block in self._traverse(start, dfs=dfs):
                yield cur_block

    def heads(self, start=None, reverse=False, dfs=False):
        """
        Iterate over instructions in function blocks.

        :param int start: optional address to start iterating from.
        :param bool reverse: iterate in reverse
        :param bool dfs: If true, traversal of blocks will be depth-first.
            If false, traversal will be breadth-first.
        """
        _first_block = True
        for cur_block in self.blocks(start, reverse=reverse, dfs=dfs):
            if start and _first_block:
                heads = cur_block.heads(start, reverse=reverse)
            else:
                heads = cur_block.heads(reverse=reverse)

            _first_block = False

            for head in heads:
                yield head

    def find_block(self, ea):
        """
        Locate a BasicBlock which contains the specified ea

        :param int ea: ea of interest

        :return: CustomBasicBlock object or None if not found.
        :rtype: CustomBasicBlock
        """
        for block in self:
            if ea in block:
                return block

    def _paths_to_ea(self, ea, cur_block, visited=None, cur_path=None):
        """
        Recursive DFS traversal of graph which yields a path to EA.

        :param int ea: ea of interesting
        :param CustomBasicBlock cur_block: current block in graph
        :param set visited: set of blocks already visited
        :param list cur_path: a list of blocks on the current path

        :yield list: current path
        """
        cur_path = cur_path or []

        # Initialize our visited set of blocks
        if visited is None:
            visited = set()

        # Mark the current block as visited and add it to the current path
        visited.add(cur_block.start_ea)
        cur_path.append(cur_block)
        # We've found our block, so yield the current path
        if ea in cur_block:
            yield copy(cur_path)

        # Continue traversing
        for block in cur_block.succs():
            if block.start_ea in visited:
                continue

            for path in self._paths_to_ea(ea, block, visited, cur_path):
                yield path

        # Remove the current block from the path and visited so it is included in subsequent paths
        cur_path.pop()
        visited.remove(cur_block.start_ea)

    def paths_to_ea(self, ea):
        """
        Yield a list which contains all the blocks on a path from the function entry point to the block
        containing the specified ea.  Raises ValueError if specified EA is not within the current function.

        :param int ea: ea of interest

        :yield list: list of BasicBlocks residing on a given path to EA
        """
        # make sure the specified ea is within the function
        if not (self.f.start_ea <= ea < self.f.end_ea):
            raise ValueError

        for path in self._paths_to_ea(ea, self[0]):
            yield path

    def get_paths(self, ea):
        """
        Given an EA, iterate over the paths to the EA.

        For usage example, see function_tracer.trace in function_tracer.py

        WARNING:
        DO NOT WRAP THIS GENERATOR IN list()!!!  This generator will iterate all possible paths to the node containing
        the specified EA.  On functions containing large numbers of jumps, the number of paths grows exponentially and
        you WILL hit memory exhaustion limits, extremely slow run times, etc.  Use extremely conservative constraints
        when iterating.  Nodes containing up to at least 32,768 paths are computed in a reasonably sane amount of time,
        though it probably doesn't make much sense to check this many paths for the data you are looking for.

        :param int ea: EA of interest

        :yield: a path to the object
        """
        # Obtain the block containing the EA of interest
        block = self.find_block(ea)

        # If block not found, then there are no paths to it.
        if not block:
            logger.debug("Unable to find block with ea: 0x{:08X}".format(ea))
            return

        for path_node in block.paths():
            yield path_node

    def _getitem(self, index):
        """
        Override the idaapi.FlowChart._getitem function to return our CustomBasicBlock
        type instead.
        """
        return CustomBasicBlock(index, self._q[index], self)



"""
Function tracing utility which will trace a function to a given EA and provide the value stored in a given operand.idaapi.__struct_unpack_table

This will essentially emulate, on a very primitive level, a x86(_64) cpu during execution and will maintain the stack,
registers, etc.
"""

import logging
from builtins import object
from typing import Tuple, List, Generator

import idaapi

from . import builtin_funcs
from .cpu_context import ProcessorContext
from .flowchart import FlowChart

logger = logging.getLogger(__name__)


class FunctionTracer(object):
    """
    This class should be instantiated by the user in order to trace a function.  Ideally, the object will be stored
    for future use should the caller expect the function traced to be analyzed multiple times.

    To use, the caller will supply the function address (or any address within the function), the address where the
    value of interest is set, the operand containing the value of interest, and optionally a list of arguments passed
    to the function.  The list of arguments will be used to set argument values for a more complete assessment of the
    function.

    The FunctionTracer object will utilize the IDAPython FlowChart object, with custom BasicBlocks.  The FlowChart
    object is utilized to efficiently grab all the instructions which are known to be executed sequentially, as well
    as ensure that the function is traced using appropriate code paths to the interesting ea.  A "snapshot" of the
    CPU register state and stack will be copied and stored for each custom BasicBlock.  The idea behind storing this
    "snapshot" is that if the function needs to be traced multiple times, the "snapshot" data can be utilized to
    initialize the state of execution when a previously unanalyzed block is being inspected rather than having to
    completely reanalyze the entire function again.
    """

    def __init__(self, func_ea):
        """
        :param func_ea: any address in the function of interest
        """
        # FIXME: Importing here to prevent cyclic imports.
        from kordesii.utils import decoderutils

        self.func_obj = decoderutils.SuperFunc_t(func_ea)
        self.func_ea = self.func_obj.start_ea
        # Create the graph object of the function
        self.flowchart = FlowChart(self.func_ea)
        self._hooks = {}

    def __repr__(self):
        return "<FunctionTracer for function at 0x{:X}>".format(self.func_ea)

    def iter_context_at(self, ea, depth=0, exhaustive=True) -> Generator[ProcessorContext, None, None]:
        """
        Iterate over cpu context for instructions up to, but not including, a given ea.

        >>> ea = 0x1001b9ad
        >>> ft = FunctionTracer(ea)
        >>> for cpu_context in ft.iter_context_at(ea):
        >>>     print cpu_context

        :param int ea: ea of interest
        :param int depth: Number of calls up the stack to pull context from.
            (defaults to 0, meaning a empty context will be generate at the top of the current function.)
        :param bool exhaustive: If true, all paths for each depth is processed
            if false, only the first path for each depth is processed.
            (defaults to exhaustive)

        :yield: cpu_context or None (if ea is the function's first address)
        """
        if depth < 0:
            raise ValueError("Depth must be a positive integer.")

        if ea not in self.func_obj:
            raise ValueError("Address 0x{:08X} not within function at 0x{:08X}".format(ea, self.func_obj.start_ea))

        # Obtaining the context consists of tracing up to, but not including ea, unless ea is the first instruction.

        for path_block in self.flowchart.get_paths(ea):
            with builtin_funcs.hooks(self._hooks):
                if not depth:
                    yield path_block.cpu_context(ea)
                    continue

                # Pull contexts from caller functions first.
                yielded = False
                for call_ea in self.func_obj.xrefs_to:
                    if call_ea in self.func_obj:
                        logger.debug("Ignoring recursive function call at 0x{:08X}".format(call_ea))
                        continue
                    tracer = get_tracer(call_ea, None)
                    if tracer:
                        for context in tracer.iter_context_at(call_ea, depth=depth - 1, exhaustive=exhaustive):
                            # increase the sp to account for the return address that gets pushed
                            # onto the stack so that we are aligned correctly.
                            context.sp -= context.byteness

                            # yield a context containing the caller executed first.
                            yield path_block.cpu_context(ea, init_context=context)
                            yielded = True
                # If we didn't yield, then we hit a function that has no callers or valid contexts.
                if not yielded:
                    yield path_block.cpu_context(ea)

            # break and don't process other paths if not exhaustive.
            if not exhaustive:
                break

    def context_at(self, ea, depth=0):
        """
        Obtain a cpu context for instructions up to, but not including, a given ea.

        :param int ea: ea of interest
        :param int depth: Number of calls up the stack to pull context from.
            (defaults to 0, meaning a empty context will be generate at the top of the current function.)

        :return: cpu_context or None
        :rtype: ProcessorContext
        """
        for ctx in self.iter_context_at(ea, depth=depth):
            return ctx

    def iter_operand_value(self, ea, index, depth=0, exhaustive=True):
        """
        Trace the function to the specified ea and yield all possible values for the operand.
        This is a helper wrapper for extracting the context and then retrieving either
        the memory address or raw value from the operand.

        NOTE: We are using "get_operand_value" to help show this is the equivalent to
        idc.get_operand_value() but pulls from emulated data.


        >>> ft = FunctionTracer(0x1001b9a0)
        >>> for ctx, val in ft.iter_operand_value(0x1001b9ad, 0):
        >>>     print "Val for opnd0 at 0x1001b9ad = 0x{:x}".format(val)

        :param int ea: address to trace to
        :param int index: the operand of interest (0 - first operand, 1 - second operand, ...)
        :param int depth: Number of calls up the stack to pull context from.
            (defaults to 0, meaning a empty context will be generate at the top of the current function.)
        :param bool exhaustive: If true, all paths for each depth is processed
            if false, only the first path for each depth is processed.
            (defaults to exhaustive)

        :yield tuple: (context at ea, operand value)
        """
        values = set()

        # Iterate all the nodes to obtain the CPU context
        for cpu_context in self.iter_context_at(ea, depth=depth, exhaustive=exhaustive):
            operand = cpu_context.operands[index]
            # Pass memory address if there is one, otherwise pass the value.
            value = operand.addr or operand.value
            # value = cpu_context.get_operand_value(index, data_size, data_type=data_type, ip=ea)
            # Prevent returning multiple values which are the same....
            if value in values:
                logger.debug("trace :: value 0x{:X} already returned.".format(value))
                continue

            values.add(value)
            yield cpu_context, value

    def get_operand_value(self, ea, index, depth=0):
        """
        Trace the function to the specified ea and return the value for the specified operand.
        This is a helper wrapper for extracting the context and then retrieving either
        the memory address or raw value from the operand.

        NOTE: We are using "get_operand_value" to help show this is the equivalent to
        idc.get_operand_value() but pulls from emulated data.

        >>> ft = FunctionTracer(0x1001b9a0)
        >>> val = ft.get_operand_value(0x1001b9ad, 0)
        >>> print "Val for opnd0 at 0x1001b9ad = 0x{:x}".format(val)

        :param int ea: address to trace to
        :param int opnd: the operand of interest (0 - first operand, 1 - second operand, ...)
        :param int depth: Number of calls up the stack to pull context from.
            (defaults to 0, meaning a empty context will be generate at the top of the current function.)

        :returns tuple: (context at ea, operand value)
        """
        for cpu_context, value in self.iter_operand_value(ea, index, depth=depth):
            return cpu_context, value

    def iter_function_args(self, ea, depth=0, exhaustive=True, num_args=None):
        """
        Given the EA of a function call, attempt to determine the number of arguments passed to the function and
        return those values to the caller.  Additionally, give back the context as well since it may be useful.

        >>> ft = FunctionTracer(0x1001b9a0)
        >>> call_addr = 0x1001b9d0
        >>> for context, args in ft.iter_function_args(call_addr):
        >>>     print "Args for call at 0x{:X}: {}".format(call_addr, ", ".join(args))

        :param int ea: address containing the function call of interest
        :param int depth: Number of calls up the stack to pull context from.
            (defaults to 0, meaning a empty context will be generate at the top of the current function.)
        :param bool exhaustive: If true, all paths for each depth is processed
            if false, only the first path for each depth is processed.
            (defaults to exhaustive)
        :param int num_args: Force a specific number of arguments.
            If not provided, number of arguments is determined by the disassembler.
            Extra arguments not defined by the disassembler are assumed to be 'int' type.

        :yield tuple: (context at ea, list of function parameters passed to called function in order)
        """
        # Iterate all the paths leading up to ea
        for cpu_context in self.iter_context_at(ea, depth=depth, exhaustive=exhaustive):
            func_ea = cpu_context.operands[0].value
            yield cpu_context, cpu_context.get_function_args(func_ea, num_args=num_args)

    def get_function_args(self, ea, depth=0, num_args=None):
        """
        Simply calls iter_function_args with the provided ea and returns the first set of arguments.

        >>> ft = FunctionTracer(0x1001b9a0)
        >>> call_addr = 0x1001b9d0
        >>> context, args = ft.get_function_args(call_addr):
        >>> print "Args for call at 0x{:X}: {}".format(call_addr, ", ".join(args))

        :param int ea: address containing the function call of interest
        :param int depth: Number of calls up the stack to pull context from.
            (defaults to 0, meaning a empty context will be generate at the top of the current function.)
        :param int num_args: Force a specific number of arguments.
            If not provided, number of arguments is determined by the disassembler.
            Extra arguments not defined by the disassembler are assumed to be 'int' type.

        :return tuple: (context at ea, list of function parameters passed to called function in order)
        :rtype: Tuple[ProcessorContext, List]
        """
        for cpu_context, args in self.iter_function_args(ea, depth=depth, num_args=num_args):
            return cpu_context, args

    def hook(self, name_or_start_ea, func):
        """
        Hooks the given name with a custom user defined function.
        Useful for emulating the effects of a function or reporting data.

        :param name_or_start_ea: Name or starting address of the function to hook (e.g. 'base64')
        :param func: Function to run while emulating a call to that function.
            Function must accept 3 arguments: cpu_context, func_name, and func_args and
            may return a value to be set to the rax register (or equivalent)
        :return:
        """
        if isinstance(name_or_start_ea, str):
            name_or_start_ea = name_or_start_ea.lower()

        self._hooks[name_or_start_ea] = func

    def clear_hooks(self):
        """
        Clears all previously set hooks.
        """
        self._hooks = {}


class TracerCache(object):
    """
    Class containing a cache of all tracer objects that are used within a decoder to avoid having to do all the
    reinitialization work on subsequent uses.
    """

    def __init__(self):
        self._tracers = {}
        self._hooks = {}

    def get(self, ea, default="NOTSET"):
        # type: (int, object) -> FunctionTracer
        """
        Get either an existing tracer for a provided EA, or create a new one.

        :param int ea: address within a function for which to retrieve a tracer
        :param default: What to return, if anything, when a function cannot be created for EA
        :return: FunctionTracer object or default
        :raises: KeyError
        """
        func = idaapi.get_func(ea)
        if not func:
            if default != "NOTSET":
                return default
            else:
                raise KeyError("Unable to create tracer")

        try:
            return self._tracers[func.start_ea]
        except KeyError:
            tracer = FunctionTracer(ea)
            for name_or_start_ea, func_hook in list(self._hooks.items()):
                tracer.hook(name_or_start_ea, func_hook)
            self._tracers[func.start_ea] = tracer
            return tracer

    def hook(self, name_or_start_ea, func):
        """
        Hooks the given name with a custom user defined function.
        Useful for emulating the effects of a function or reporting data.
        All tracers produced will contain this hook when used here.

        :param name_or_start_ea: Name or starting address of the function to hook (e.g. 'base64')
        :param func: Function to run while emulating a call to that function.
            Function must accept 3 arguments: cpu_context, func_name, and func_args and
            may return a value to be set to the rax register (or equivalent)
        :return:
        """
        if isinstance(name_or_start_ea, str):
            name_or_start_ea = name_or_start_ea.lower()

        # Hook already created tracers.
        for tracer in list(self._tracers.values()):
            tracer.hook(name_or_start_ea, func)

        # Save hook for future tracers.
        self._hooks[name_or_start_ea] = func

    def clear_hooks(self):
        """Clears all previously set hooks."""
        self._hooks = {}
        for tracer in list(self._tracers.values()):
            tracer.clear_hooks()


# Create a global TracerCache that can be used throughout.
_tracer_cache = TracerCache()
get_tracer = _tracer_cache.get
hook_tracers = _tracer_cache.hook
clear_hooks = _tracer_cache.clear_hooks



"""
Implements the "hardware" for tracing a function.

Will perform the instructions and updates CPU registers, stack information, etc.

WARNING:
    Do NOT rely on the flags registers being correct.  There are places were flags are NOT being updated when they
    should, and the very fact that CALL instructions are skipped could cause flags to be incorrect.
"""
from copy import deepcopy
import collections
import logging

import idaapi
import idc
import ida_struct
import ida_ua

from kordesii.utils.function_tracing import utils
from kordesii.utils.function_tracing.constants import *
from kordesii.utils.function_tracing.memory import Memory
from kordesii.utils.function_tracing.variables import VariableMap
from kordesii.utils.function_tracing.operands import Operand, OperandLite
from kordesii.utils.function_tracing.functions import FunctionSignature


logger = logging.getLogger(__name__)


class JccContext(object):
    """
    Stores information pertaining to a Jcc instruction encountered when tracing.

    When a Jcc instruction is encountered, several pieces of information inherently need to be tracked since
    we are blindly taking every branch to ensure we get all possible data at any given address.  It turns out
    we need to know the target of the Jcc instruction for the condition as emulated
    (condition_target_ea).  We also need to know the value of the branch we would NOT have taken (at least as
    best of a guess as we can make in some cases) and where that value would have been set.  In order to
    calculate the value, we need to know what kind of test instruction was used, so that mnem is tracked as well.When
    we trace our condition_target_ea branch, we need not modify the context.  Whenever we trace the alternative branch,
    we'll need to modify the context as specified.
    """

    def __init__(self):
        self.condition_target_ea = None  # The branch actually taken
        self.alt_branch_data_dst = None  # The location which was tested (typically opnd 0 of the condition test)
        self.alt_branch_data = None  # The data stored in _alt_branc_data_dst
        self.flag_opnds = {}  # Dictionary containing the operands at a particular instruction which set
        # specific flags.  Dictionary is keyed on flag registery names.

    def __deepcopy__(self, memo):
        copy = JccContext()
        copy.condition_target_ea = self.condition_target_ea
        copy.alt_branch_data_dst = self.alt_branch_data_dst
        copy.alt_branch_data = self.alt_branch_data
        copy.flag_opnds = {flag: list(operands) for flag, operands in list(self.flag_opnds.items())}
        return copy

    def update_flag_opnds(self, flags, opnds):
        """
        Set the operands which last changed the specified flags.

        :param flags: list of flags which were modified utilizing the supplied opnds
        :param opnds: list of operands (instance of Operand) at the instruction which modified the flags
        """
        for flag in flags:
            # Converting Operand classes to OperandLite classes to help speed up deepcopies.
            self.flag_opnds[flag] = [OperandLite(opnd.ip, opnd.idx, opnd.text, opnd.value) for opnd in opnds]

    def get_flag_opnds(self, flags):
        """
        Extracts all the operands of for the list of flags and reduces the set.  However, since the operands
        need to remain in order, we can't use set operations.  In all actuality, assuming our code is correct and
        the compiler isn't doing something funky, any more than 1 flag should really just be a duplicate list.

        :param flags: list of flags for which to extract operands
        :return: list of operands which were utilized in the instruction that modified the requested flags
        """
        # TODO: Is there a better way to do this?
        operands = []
        for flag in flags:
            for operand in self.flag_opnds.get(flag, []):
                if operand not in operands:
                    operands.append(operand)

        return operands

    def is_alt_branch(self, ip):
        """
        Test our IP against the branch information to determine if we are in the branch that would have been
        emulated or in the alternate branch.
        """
        return self.condition_target_ea and self.condition_target_ea != ip


# TODO: Create architecture specific Context types.
class ProcessorContext(object):
    """
    Stores the context of the processor during execution.

    :param registers: Instance of an initialized RegisterMap object used to store register values
        for the given architecture.
    :param str instruction_pointer: Name of the register used to point to the current instruction
        being currently executed or to-be executed.
    :param [str] stack_registers: List of register names used for handling the stack.
    """

    # Must be set by inherited classes.
    ARCH_NAME = None  # Name of architecture as reported by disassembler.
    OPCODES = {}  # Map of opcode mnemonics to functions that emulate them.

    def __init__(self, registers, instruction_pointer, stack_pointer, stack_registers=None):
        self.registers = registers
        self.jcccontext = JccContext()
        self.memory = Memory()
        self.func_calls = {}  # Keeps track of function calls.
        self.executed_instructions = []  # Keeps track of the instructions that have been executed.
        self.memory_copies = collections.defaultdict(list)  # Keeps track of memory moves.
        self.bitness = utils.get_bits()
        self.byteness = self.bitness // 8
        self.stack_registers = stack_registers or []
        self.variables = VariableMap(self)
        self._sp = stack_pointer
        self._ip = instruction_pointer

    @classmethod
    def from_arch(cls, arch_name=None):
        """
        Factory method for initializing a ProcessorContext based on detected architecture.

        :param arch_name: Name of architecture to initializes (according to the disassembler)
                          Architecture is automatically detected if not provided.

        :raises NotImplementedError: If architecture is not supported.
        """
        # Pull from disassembler if not provided.
        if not arch_name:
            info = idaapi.get_inf_structure()
            arch_name = info.procName

        for subclass in cls.__subclasses__():
            if subclass.ARCH_NAME == arch_name:
                return subclass()  # Subclasses shouldn't have any initialization parameters.
        raise NotImplementedError("Architecture not supported: {}".format(arch_name))

    def __deepcopy__(self, memo):
        """Implementing our own deepcopy to improve speed."""
        # Create class, but avoid calling __init__()
        # so we don't trigger the unnecessary initialization of Memory and JccContext
        klass = self.__class__
        copy = klass.__new__(klass)
        memo[id(self)] = copy

        copy.registers = deepcopy(self.registers, memo)
        copy.jcccontext = deepcopy(self.jcccontext, memo)
        copy.memory = deepcopy(self.memory, memo)
        copy.variables = deepcopy(self.variables, memo)
        copy.func_calls = dict(self.func_calls)
        copy.executed_instructions = list(self.executed_instructions)
        copy.memory_copies = self.memory_copies.copy()
        copy.bitness = self.bitness
        copy.byteness = self.byteness
        copy.stack_registers = self.stack_registers
        copy._sp = self._sp
        copy._ip = self._ip

        return copy

    @property
    def ip(self):
        """Alias for retrieving instruction pointer."""
        return self.registers[self._ip]

    @ip.setter
    def ip(self, value):
        """Alias for setting instruction pointer."""
        self.registers[self._ip] = value

    @property
    def sp(self):
        """Alias for retrieving stack pointer."""
        return self.registers[self._sp]

    @sp.setter
    def sp(self, value):
        """Alias for setting stack pointer."""
        self.registers[self._sp] = value

    @property
    def prev_instruction(self):
        """That last instruction that was executed or None if no instructions have been executed."""
        if self.executed_instructions:
            return self.executed_instructions[-1]
        else:
            return None

    def execute(self, ip=None):
        """
        "Execute" the instruction at IP and store results in the context.
        The RIP/EIP register will be set to the value supplied in IP so that it is
        correct.

        :param ip: instruction address to execute (defaults to currently set ip)
        """
        if not ip:
            ip = self.ip

        # Set instruction pointer to where we are currently executing.
        self.ip = ip

        # Determine if a rep* instruction and add termination condition.
        term_condition = None
        if idc.get_wide_byte(ip) in (0xF2, 0xF3):
            insn = idc.GetDisasm(ip)  # IDA pro never has operands for rep opcodes.
            if insn.startswith("rep "):
                term_condition = lambda: self.registers.ecx == 0
            elif insn.startswith(("repe ", "repz ")):
                term_condition = lambda: self.registers.ecx == 0 or self.registers.zf == 0
            elif insn.startswith(("repne ", "repnz ")):
                term_condition = lambda: self.registers.ecx == 0 or self.registers.zf == 1

        # Emulate instruction.
        mnem = idc.print_insn_mnem(ip)
        operands = self.operands
        instruction = self.OPCODES.get(mnem)
        if instruction:
            try:
                if term_condition:
                    # As a safety measure, don't allow rep instructions to surpass
                    # our max memory read limit.
                    if self.registers.ecx > self.memory.MAX_MEM_READ:
                        logger.warning(
                            "0x{:08X} :: Emulation attempted to read {} instruction {} times. "
                            "Ignoring instruction.".format(ip, mnem, self.registers.ecx)
                        )
                    else:
                        logger.debug("Emulating {} instruction {} times.".format(mnem, self.registers.ecx))
                        while not term_condition():
                            instruction(self, ip, mnem, operands)
                            self.registers.ecx -= 1
                else:
                    instruction(self, ip, mnem, operands)
            except Exception:
                logger.exception("Failed to execute address 0x{:X}: {}".format(ip, idc.GetDisasm(ip)))
        else:
            logger.debug("{} instruction not implemented.".format(mnem))

        # Record executed instruction.
        self.executed_instructions.append(ip)

        # After execution, set instruction pointer to next instruction assuming
        # standard code flow and if no jump was made.
        if self.ip == ip:
            self.ip = idc.next_head(ip)

    def get_call_history(self, func_name):
        """
        Returns the call history for a specific function name.

        :returns: List of tulples containing: (ea of call, list of function arguments)
        """
        return [(ea, args) for ea, (_func_name, args) in list(self.func_calls.items()) if _func_name == func_name]

    def prep_for_branch(self, bb_start_ea):
        """
        Modify this current context in preparation for a specific path.
        """
        if self.jcccontext.is_alt_branch(bb_start_ea):
            logger.debug("Modifying context for branch at 0x{:X}".format(bb_start_ea))
            # Set the destination operand relative to the current context
            # to a valid value that makes this branch true.
            dst_opnd = self.jcccontext.alt_branch_data_dst
            dst_opnd = self.get_operands(ip=dst_opnd.ip)[dst_opnd.idx]
            dst_opnd.value = self.jcccontext.alt_branch_data

        self.jcccontext = JccContext()

    def get_operands(self, ip=None):
        """
        Gets the Operand objects of all operands in the current instruction and returns them in a list.

        :param int ip: location of instruction pointer to pull operands from (defaults to current rip in context)

        :return: list of Operand objects
        """
        if ip is None:
            ip = self.ip

        operands = []
        cmd = ida_ua.insn_t()
        # NOTE: We can't trust the instruction length returned by decode_ins.
        ida_ua.decode_insn(cmd, ip)
        for idx, op in enumerate(cmd.ops):
            operand = Operand(self, ip, idx)
            # IDA will sometimes create hidden or "fake" operands.
            # These are there to represent things like an implicit EAX register.
            # To help avoid confusion to the opcode developer, these fake operands will not be included.
            if not operand.is_hidden:
                operands.append(operand)

            if operand.is_void:
                break  # no more operands

        return operands

    @property
    def operands(self):
        return self.get_operands()

    def reg_read(self, reg):
        """
        Read a register value

        >>> cpu_context = ProcessorContext()
        >>> cpu_context.reg_read("EIP")

        :param str reg: register name to be read

        :return int: value contained in specified register as int
        """
        return self.registers[reg]

    def reg_write(self, reg, val):
        """
        Write a register value

        :param str reg: register name to be written

        :param int val: value to be written to register as an int of width of the register (will be truncated as necessary)
        """
        self.registers[reg] = val

    def mem_alloc(self, size):
        """
        Allocates heap region with size number of bytes.

        :param size: Number of bytes to allocate.
        :return: starting address of allocated memory.
        """
        return self.memory.alloc(size)

    def mem_realloc(self, address, size):
        """
        Reallocates heap region with size number of bytes.

        :param address: base address to reallocate.
        :param size: Number of bytes to allocate.
        :return: address of the reallocated memory block.
        """
        new_address = self.memory.realloc(address, size)
        # Record a memory copy if pointer has changed.
        if new_address != address:
            self.memory_copies[self.ip].append((address, new_address, size))
        return new_address

    def mem_copy(self, src, dst, size):
        """
        Copy data from src address to dst address
        (Use this over mem_read/mem_write in order to allow the context to keep track of memory pointer history.)

        :param src: Source address
        :param dst: Destination address
        :param size: Number of bytes to copy over.
        :return:
        """
        self.memory_copies[self.ip].append((src, dst, size))
        self.mem_write(dst, self.mem_read(src, size))

    def get_pointer_history(self, ea):
        """
        Retrieves the history of a specific pointer.
        :param ea: Pointer to start with.
        :return: list of tuples containing (address of the memory copy, source pointer)
            - sorted by earliest to latest incarnation of the pointer. (not including itself)
        """
        history = []
        for ip, copies in sorted(list(self.memory_copies.items()), reverse=True):
            for src, dst, size in sorted(copies, reverse=True):
                if dst == ea:
                    history.append((ip, src))
                    ea = src
        history.reverse()
        return history

    def get_original_location(self, addr):
        """
        Retrieves the original location for a given address by looking through it's pointer history.

        :param addr: address of interest

        :return: a tuple containing:
            - instruction pointer where the original location was first copied
                or None if given address is already loaded or the original location could not be found.
            - either a loaded address, a tuple containing (frame_id, stack_offset) for a stack variable,
                or None if the original location could not be found.
        """
        # TODO: Consider refactoring.

        # Pull either the first seen loaded address or last seen stack variable.
        if idc.is_loaded(addr):
            return None, addr
        ip = None

        var = self.variables.get(addr, None)
        for ip, ea in reversed(self.get_pointer_history(addr)):
            if idc.is_loaded(ea):
                return ip, ea
            var = self.variables.get(ea, var)

        if var and var.is_stack:
            return ip, (var.frame_id, var.stack_offset)
        else:
            return ip, None

    def mem_read(self, address, size):
        """
        Read memory at the specified address of size size

        :param int address: address to read memory from
        :param int size: size of data to be read
        :return bytes: read data as bytes
        """
        return self.memory.read(address, size)

    def mem_write(self, address, data):
        """
        Write content contained in data to specified address

        :param int address: address to write data at
        :param bytes data: data to be written as bytes
        """
        self.memory.write(address, data)

    def mem_find(self, value, start=0, end=None):
        return self.memory.find(value, start=start, end=end)

    def mem_find_in_segment(self, value, seg_name_or_ea):
        return self.memory.find_in_segment(value, seg_name_or_ea)

    def mem_find_in_heap(self, value):
        return self.memory.find_in_heap(value)

    def read_data(self, addr, size=None, data_type=None):
        """
        Reads memory at the specified address, of the specified size and convert
        the resulting data into the specified type.

        :param int addr: address to read data from
        :param int size: size of data to read
        :param data_type: type of data to be extracted
            (default to byte string is size provided or C string if not)
        """
        if not data_type:
            data_type = STRING if size is None else BYTE_STRING
        if size is None:
            size = 0

        if data_type == STRING:
            null_offset = self.memory.find(b"\0", start=addr)
            # It should always eventually find a null since unmapped pages
            # are all null. If we get -1 we have a bug.
            assert null_offset != -1, "Unable to find a null character!"
            return self.memory.read(addr, null_offset - addr)

        elif data_type == WIDE_STRING:
            # Step by 2 bytes to find 2 nulls on an even alignment.
            # (This helps prevent the need to take endianness into account.)
            null_offset = addr
            while self.memory.read(null_offset, 2) != b"\0\0":
                null_offset += 2

            return self.memory.read(addr, null_offset - addr)

        elif data_type == BYTE_STRING:
            return self.memory.read(addr, size)

        elif data_type == BYTE:
            return utils.struct_unpack(self.mem_read(addr, 1))

        elif data_type == WORD:
            return utils.struct_unpack(self.mem_read(addr, 2))

        elif data_type == DWORD:
            return utils.struct_unpack(self.mem_read(addr, 4))

        elif data_type == QWORD:
            return utils.struct_unpack(self.mem_read(addr, 8))

        raise ValueError("Invalid data_type: {!r}".format(data_type))

    def get_function_signature(self, func_ea=None, force=False) -> FunctionSignature:
        """
        Returns the function signature of the given func_ea with argument values pulled
        from this context.

        :param int func_ea: address of the function to pull signature from.
            The first operand is used if not provided. (helpful for a "call" instruction)
        :param bool force: Whether to force a function signature using cdecl calling
            convention and no arguments if we fail to generate the signature.
            (Useful when trying to declare a function that was dynamically created in a register)
        :return: FunctionSignature object

        :raises RuntimeError: If a function signature could not be created from given ea.
        """
        # If func_ea is not given, assume we are using the first operand from a call instruction.
        if not func_ea:
            operand = self.operands[0]
            # function pointer can be a memory reference or immediate.
            func_ea = operand.addr or operand.value

        try:
            return FunctionSignature(self, func_ea)
        except RuntimeError as e:
            # If we fail to get a function signature but force is set, set the type to
            # cdecl with no arguments.
            if force:
                logger.warning(
                    "Failed to create function signature at 0x{:0X} with error: {}\n"
                    "Forcing signature with assumed cdecl calling convention.".format(func_ea, e)
                )
                idc.SetType(func_ea, "int __cdecl no_name();")
                return FunctionSignature(self, func_ea)
            else:
                raise

    def get_function_args(self, func_ea=None, num_args=None):
        """
        Returns the function argument values for this context based on the
        given function.

        >>> cpu_context = ProcessorContext()
        >>> args = cpu_context.get_function_args(0x180011772)

        :param int func_ea: Ea of the function to pull a signature from.
        :param int num_args: Force a specific number of arguments.
            If not provided, number of arguments is determined by the disassembler.
            Extra arguments not defined by the disassembler are assumed to be 'int' type.
            Use get_function_signature() and adjust the FunctionSignature manually
            if more customization is needed.
            (NOTE: The function signature will be forced on failure if this is set.)

        :returns: list of function arguments
        """
        func_sig = self.get_function_signature(func_ea, force=num_args is not None)

        if num_args is not None:
            if num_args < 0:
                raise ValueError("num_args is negative")
            arg_types = func_sig.arg_types
            if len(arg_types) > num_args:
                func_sig.arg_types = arg_types[:num_args]
            elif len(arg_types) < num_args:
                func_sig.arg_types = arg_types + ("int",) * (num_args - len(arg_types))

        return [arg.value for arg in func_sig.args]



"""
Interface for memory management.
"""

import collections
import logging
from copy import deepcopy

import ida_bytes
import ida_segment
import idc

from kordesii.utils.function_tracing.utils import get_bits

logger = logging.getLogger(__name__)


class PageMap(collections.defaultdict):
    """
    Dictionary of page indexes to pages.

    Creates a new page when missing.
    New pages uses the bytes from the IDB if in a segment.
    Segments pages will be mapped, but data retrieval will be delayed until the page
    is requested. (Helps to avoid unnecessary processing of large unused data segments.)
    """

    PAGE_SIZE = 0x1000

    # Cache of segment pages.
    # Used to prevent multiple calls to pull data from the IDB.
    _segment_cache = {}

    def __init__(self, map_segments=True):
        # Setting default_factory to None, because we have overwritten it in __missing__()
        super(PageMap, self).__init__(None)
        if map_segments:
            self.map_segments()

    def __deepcopy__(self, memo):
        copy = PageMap(map_segments=False)
        memo[id(self)] = copy
        copy.update({index: (page[:] if page is not None else None) for index, page in self.items()})
        return copy

    def __missing__(self, page_index):
        """
        Creates a new page when index first encountered.

        :return: page
        :rtype: bytearray
        """
        ret = self[page_index] = self._new_page(page_index)
        return ret

    def __getitem__(self, page_index):
        try:
            page = super(PageMap, self).__getitem__(page_index)
        except KeyError:
            return self.__missing__(page_index)

        # If page is None, that means this was set for delayed retrieval.
        # Retrieve page now that it is being requested.
        if page is None:
            return self.__missing__(page_index)

        return page

    def _is_delayed(self, page_index):
        """Determines if page is set for delayed retrieval."""
        return page_index in self and super(PageMap, self).__getitem__(page_index) is None

    def map_segments(self):
        """Sets segment pages for delayed retrieval"""
        for n in range(ida_segment.get_segm_qty()):
            seg = ida_segment.getnseg(n)
            if seg:
                for page_index in range(seg.start_ea >> 12, ((seg.end_ea - 1) >> 12) + 1):
                    self[page_index] = None

    @staticmethod
    def _obtain_bytes(start, end):
        """
        Obtain bytes efficiently, sets non-loaded bytes to \x00

        :param int start: starting address
        :param int end: ending address

        :return bytearray: bytearray containing bytes within range
        """
        # Reconstruct the segment, account for bytes which are not loaded.
        bytes_range = range(start, end)  # a range from start -> end
        return bytearray(ida_bytes.get_wide_byte(i) if idc.is_loaded(i) else 0 for i in bytes_range)

    def _new_page(self, page_index):
        """
        Creates a new page based on index.

        :return: page
        :rtype: bytearray
        """
        if page_index in self._segment_cache:
            return self._segment_cache[page_index][:]

        start_ea = page_index * self.PAGE_SIZE
        end_ea = start_ea + self.PAGE_SIZE

        # If page was set for delayed retrieval it is coming from segment data, so pull from IDB.
        # Update this check if ever use delayed retrieval for non-segment data.
        if self._is_delayed(page_index):
            logger.debug("Reading segment data 0x{:X} -> 0x{:X} from IDB".format(start_ea, end_ea))
            page = self._obtain_bytes(start_ea, end_ea)
            self._segment_cache[page_index] = page[:]  # cache page first
            return page

        # If range is not in a segment, provide a page of all zeros.
        return bytearray(self.PAGE_SIZE)

    def peek(self, page_index):
        """
        Returns the page for the given page index.
        If page doesn't exist, it creates the page but doesn't set it in the map.

        .. warning:: If you are using this you shouldn't try to modify the page since its
            effects may not propagate to the map.
            (ie. this is a read-only copy)

        :param page_index:
        :return: page
        :rtype: bytearray
        """
        if page_index in self and not self._is_delayed(page_index):
            return self[page_index]
        return self._new_page(page_index)


def clear_cache():
    """
    Clears the internal cache of segment bytes.
    Calling this will be necessary if you have patched in new bytes into the IDB.
    """
    PageMap._segment_cache = {}


class Memory(object):
    """
    Class which implements the CPU memory controller backed by the segment data in the input file.

    This class provides a read() and write() function for CPU emulation.
    If a memory address has not been written to, null bytes will be returned.
    """

    PAGE_SIZE = PageMap.PAGE_SIZE

    HEAP_BASE = idc.get_inf_attr(idc.INF_MAX_EA)
    # Slack space between heap allocations.
    HEAP_SLACK = 0x3000

    # maximum amount of memory allowed to read/write
    # (if we are reading/writing more than ~ 268 MB we have bigger problems.)
    MAX_MEM_READ = 0x10000000
    MAX_MEM_WRITE = 0x10000000

    def __init__(self, _copying=False):
        """Initializes Memory object."""
        self._pages = PageMap(map_segments=not _copying)
        # A map of base addresses to size for heap allocations.
        self._heap_allocations = {}

    def __deepcopy__(self, memo):
        copy = Memory(_copying=True)
        memo[id(self)] = copy
        copy._pages = deepcopy(self._pages, memo)
        copy._heap_allocations = self._heap_allocations.copy()
        return copy

    @property
    def blocks(self):
        """
        Returns a list of tuples containing the base address and size for
        contiguous blocks of memory.
        """
        # Collect ranges of continuous memory.
        memory_ranges = []
        base_address = None
        size = 0
        for page_index in sorted(self._pages):
            # First page of block?
            if base_address is None:
                base_address = page_index << 12

            size += self.PAGE_SIZE

            # Found end of continuous block of memory?
            if page_index + 1 not in self._pages:
                memory_ranges.append((base_address, size))
                base_address = None
                size = 0

        # Store last block
        if base_address is not None:
            memory_ranges.append((base_address, size))

        return memory_ranges

    def __str__(self):
        """
        Print information about current memory map.
        """
        # Create text output.
        _just = 25 if get_bits() == 32 else 50
        _hex_fmt = "0x{:08X}" if get_bits() == 32 else "0x{:016X}"
        title = "{}{}{}".format("Base Address".ljust(_just), "Address Range".ljust(_just), "Size")
        memory_ranges = []
        for base_address, size in self.blocks:
            memory_ranges.append(
                "{}{}{}".format(
                    _hex_fmt.format(base_address).ljust(_just),
                    "{} - {}".format(_hex_fmt.format(base_address), _hex_fmt.format(base_address + size)).ljust(_just),
                    size,
                )
            )

        return "{}\n{}\n".format(title, "\n".join(memory_ranges))

    def is_mapped(self, address):
        return address >> 12 in self._pages

    def alloc(self, size):
        """
        Allocates heap region with size number of bytes.

        :param size: Number of bytes to allocate.
        :return: starting address of allocated memory.
        """
        # Allocate from HEAP_BASE if this is our first allocation.
        if not self._heap_allocations:
            address = self.HEAP_BASE
        # Otherwise, use the largest base address not used.
        # TODO: We may want to reuse previously freed space in the future.
        else:
            max_base_address = max(self._heap_allocations)
            heap_size = self._heap_allocations[max_base_address]
            address = max_base_address + heap_size + self.HEAP_SLACK

        # NOTE: We are just going to record that the memory as been allocated
        # but not actually trigger any data from being written. (The calls to write() will do that)
        # This helps to prevent us from wasting (real) memory if someone allocates
        # a huge amount of memory but only uses a small amount.
        self._heap_allocations[address] = size
        logger.debug("[alloc] :: Allocated {} bytes at 0x{:X}".format(size, address))
        return address

    def realloc(self, address, size):
        """
        Reallocates heap region with size number of bytes.

        :param address: base address to reallocate.
        :param size: Number of bytes to allocate.
        :return: address of the reallocated memory block.
        """
        # Passed in address should be the base address of a previously allocated memory region.
        if address not in self._heap_allocations:
            raise ValueError("0x{:X} address is not allocated.".format(address))

        previous_size = self._heap_allocations[address]

        # See if we need to relocate the heap address.
        if size > previous_size:
            for base_address in sorted(self._heap_allocations):
                if address < base_address < address + size:
                    # We need to relocate the memory block.
                    new_address = self.alloc(size)

                    # Copy over data from previous allocation.
                    # Since relocation is very rare, we will accept the loss in cycles
                    # if we end up writing emptyf data.
                    self.write(new_address, self.read(address, previous_size))

                    # Don't free the old, because the user may want to search it.
                    logger.debug("[realloc] :: Relocated 0x{:X} -> 0x{:X}".format(address, new_address))
                    return new_address

        # Otherwise we just need to adjust the size.
        if previous_size != size:
            logger.debug(
                "[realloc] :: Reallocating heap size at 0x{:X} from {} to {} bytes.".format(
                    address, previous_size, size
                )
            )
            self._heap_allocations[address] = size
        return address

    def read(self, address, size):
        """
        Reads data from given address.

        :param address: Address to read data from.
        :param size: Number of bytes to read.

        :return: byte string of read data.
        """
        if address < 0:
            raise ValueError("Address must be a positive integer. Got 0x{:08X}".format(address))
        if size < 0:
            raise ValueError("Size must be a positive integer.")
        if size > self.MAX_MEM_READ:
            logger.error(
                "[mem_read] :: Attempted to read {} bytes from 0x{:08X}. "
                "Ignoring request and reading {} bytes instead.".format(size, address, self.MAX_MEM_READ)
            )
            size = self.MAX_MEM_READ

        logger.debug("[mem_read] :: Reading {} bytes from 0x{:08X}".format(size, address))

        page_index = address >> 12
        page_offset = address & 0xFFF

        # Read data from pages.
        out = bytearray()
        while size:
            # We don't want to trigger a page creation on read().. only write().
            page = self._pages.peek(page_index)
            read_bytes = page[page_offset : page_offset + size]
            out += read_bytes
            size -= len(read_bytes)
            page_offset = 0
            page_index += 1

        return bytes(out)

    def write(self, address, data):
        """
        Writes data to given addre ss.

        :param address: Address to write data to.
        :param data: data to write
        """
        if address < 0:
            raise ValueError("Address must be a positive integer. Got 0x{:08X}".format(address))

        size = len(data)
        if size > self.MAX_MEM_WRITE:
            logger.error(
                "[mem_read] :: Attempted to write {} bytes to 0x{:08X}. "
                "Ignoring request and using first {} bytes instead.".format(size, address, self.MAX_MEM_WRITE)
            )
            data = data[: self.MAX_MEM_WRITE]

        logger.debug("[mem_write] :: Writing {} bytes to 0x{:08X}".format(len(data), address))

        page_index = address >> 12
        page_offset = address & 0xFFF

        # Write data into pages.
        while data:
            page = self._pages[page_index]
            split_index = self.PAGE_SIZE - page_offset
            to_write = data[:split_index]
            try:
                page[page_offset : page_offset + len(to_write)] = to_write
            except TypeError:
                raise TypeError("to_write: {} {}".format(type(to_write), repr(to_write)))
            data = data[split_index:]
            page_offset = 0
            page_index += 1

    def find(self, value, start=0, end=None):
        """
        Searches memory for given value.

        :param bytes value: byte string to search for
        :param int start: Starting address to start search (defaults to 0)
        :param int end: Optional ending address to end search

        :return int: address where value was located or -1 if not found

        :raises ValueError: If search value is larger than a page.
        """
        # We are not going to handle things that could expand beyond multiple pages
        if len(value) >= self.PAGE_SIZE:
            raise ValueError("Search value must be less than {} bytes, got {}".format(self.PAGE_SIZE, len(value)))

        if end and end <= start:
            raise ValueError("Ending address must be greater than starting address.")

        page_index = start >> 12
        page_offset = start & 0xFFF

        if end:
            end_page_index = end >> 12
            end_page_offset = end & 0xFFF
        else:
            end_page_index = end_page_offset = None

        page = self._pages.peek(page_index)

        # First search for the entire value in the page.
        if end and end_page_index == page_index:
            # Since we end in this page, don't attempt to read overlap.
            offset = page.find(value, page_offset, end_page_offset)
            if offset <= -1:
                return -1
            return page_index << 12 | offset

        offset = page.find(value, page_offset)
        if offset > -1:
            return page_index << 12 | offset

        # If we can't find it, try again with part of the next page attached
        # to account for data overlapping onto another page.
        _start = max(page_offset, self.PAGE_SIZE - len(value) + 1)
        _end = self.PAGE_SIZE + len(value) - 1
        if end and end_page_index == page_index + 1:
            _end = min(_end, self.PAGE_SIZE + end_page_offset)
            if _end == self.PAGE_SIZE + end_page_offset and _end <= _start:
                return -1

        next_page = self._pages.peek(page_index + 1)
        offset = (page + next_page).find(value, _start, _end)
        if offset > -1:
            return page_index << 12 | offset
        else:
            # Jump to the next mapped page to continue the search.
            # return -1 if there are no more pages beyond.
            for _page_index in sorted(self._pages):
                if _page_index > page_index:
                    # Stop searching pages if we surpass the end.
                    if end and end <= _page_index << 12:
                        return -1
                    return self.find(value, start=_page_index << 12, end=end)
            return -1

    def finditer(self, value, start=0, end=None):
        """
        Searches for all instances of value within memory.
        """
        while True:
            offset = self.find(value, start=start, end=end)
            if offset == -1:
                return
            yield offset
            start = offset + len(value)

    def find_in_segment(self, value, seg_name_or_ea):
        """
        Searches memory for given value within the range of a specific segment.

        :param bytes value: byte string to search for
        :param seg_name_or_ea: segment name or address within segment.

        :return int: address where value was located or -1 if not found
        """
        if isinstance(seg_name_or_ea, str):
            segment = ida_segment.get_segm_by_name(seg_name_or_ea)
        else:
            segment = ida_segment.getseg(seg_name_or_ea)

        return self.find(value, start=segment.start_ea, end=segment.end_ea)

    def finditer_in_segment(self, value, seg_name_or_ea):
        """
        Searches memory for given value within the range of a specific segment.

        :param bytes value: byte string to search for
        :param seg_name_or_ea: segment name or address withing segment.

        :return int: address where value was located or -1 if not found
        """
        if isinstance(seg_name_or_ea, str):
            segment = ida_segment.get_segm_by_name(seg_name_or_ea)
        else:
            segment = ida_segment.getseg(seg_name_or_ea)

        for offset in self.finditer(value, start=segment.start_ea, end=segment.end_ea):
            yield offset

    def find_in_heap(self, value):
        """
        Searches memory for given value within the allocated heap.

        :param bytes value: byte string to search for

        :return int: address where value was located or -1 if not found
        """
        return self.find(value, start=self.HEAP_BASE)

    def finditer_in_heap(self, value):
        """
        Searches memory for given value within the allocated heap.

        :param bytes value: byte string to search for

        :return int: address where value was located or -1 if not found
        """
        for offset in self.finditer(value, start=self.HEAP_BASE):
            yield offset



from copy import deepcopy
import logging

from ..registers import Register, RegisterMap


logger = logging.getLogger(__name__)


class x86_64_Registers(RegisterMap):
    """Initializes registers for x86/x64 architecture"""

    def __init__(self):
        registers = [
            Register(8, rax=0xFFFFFFFFFFFFFFFF, eax=0xFFFFFFFF, ax=0xFFFF, al=0xFF, ah=0xFF00),
            Register(8, rbx=0xFFFFFFFFFFFFFFFF, ebx=0xFFFFFFFF, bx=0xFFFF, bl=0xFF, bh=0xFF00),
            Register(8, rcx=0xFFFFFFFFFFFFFFFF, ecx=0xFFFFFFFF, cx=0xFFFF, cl=0xFF, ch=0xFF00),
            Register(8, rdx=0xFFFFFFFFFFFFFFFF, edx=0xFFFFFFFF, dx=0xFFFF, dl=0xFF, dh=0xFF00),
            Register(8, rsi=0xFFFFFFFFFFFFFFFF, esi=0xFFFFFFFF, si=0xFFFF, sil=0xFF),
            Register(8, rdi=0xFFFFFFFFFFFFFFFF, edi=0xFFFFFFFF, di=0xFFFF, dil=0xFF),
            Register(8, rbp=0xFFFFFFFFFFFFFFFF, ebp=0xFFFFFFFF, bp=0xFFFF, bpl=0xFF),
            Register(8, rsp=0xFFFFFFFFFFFFFFFF, esp=0xFFFFFFFF, sp=0xFFFF, spl=0xFF),
            Register(8, rip=0xFFFFFFFFFFFFFFFF),
            Register(8, r8=0xFFFFFFFFFFFFFFFF, r8d=0xFFFFFFFF, r8w=0xFFFF, r8b=0xFF),
            Register(8, r9=0xFFFFFFFFFFFFFFFF, r9d=0xFFFFFFFF, r9w=0xFFFF, r9b=0xFF),
            Register(8, r10=0xFFFFFFFFFFFFFFFF, r10d=0xFFFFFFFF, r10w=0xFFFF, r10b=0xFF),
            Register(8, r11=0xFFFFFFFFFFFFFFFF, r11d=0xFFFFFFFF, r11w=0xFFFF, r11b=0xFF),
            Register(8, r12=0xFFFFFFFFFFFFFFFF, r12d=0xFFFFFFFF, r12w=0xFFFF, r12b=0xFF),
            Register(8, r13=0xFFFFFFFFFFFFFFFF, r13d=0xFFFFFFFF, r13w=0xFFFF, r13b=0xFF),
            Register(8, r14=0xFFFFFFFFFFFFFFFF, r14d=0xFFFFFFFF, r14w=0xFFFF, r14b=0xFF),
            Register(8, r15=0xFFFFFFFFFFFFFFFF, r15d=0xFFFFFFFF, r15w=0xFFFF, r15b=0xFF),
            Register(2, gs=0xFFFF),
            Register(2, fs=0xFFFF),
            Register(2, es=0xFFFF),
            Register(2, ds=0xFFFF),
            Register(2, cs=0xFFFF),
            Register(2, ss=0xFFFF),
            Register(16, xmm0=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF),
            Register(16, xmm1=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF),
            Register(16, xmm2=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF),
            Register(16, xmm3=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF),
            Register(16, xmm4=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF),
            Register(16, xmm5=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF),
            Register(16, xmm6=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF),
            Register(16, xmm7=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF),
            Register(16, xmm8=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF),
            Register(16, xmm9=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF),
            Register(16, xmm10=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF),
            Register(16, xmm11=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF),
            Register(16, xmm12=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF),
            Register(16, xmm13=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF),
            Register(16, xmm14=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF),
            Register(16, xmm15=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF),
            # FLAGS register
            Register(
                4,
                **{
                    "cf": 0x1,
                    "pf": 0x4,
                    "af": 0x10,
                    "zf": 0x40,
                    "sf": 0x80,
                    "tf": 0x100,
                    "if": 0x200,
                    "df": 0x400,
                    "of": 0x800,
                    "iopl": 0x2000,
                    "nt": 0x4000,
                    "rf": 0x10000,
                    "vm": 0x20000,
                    "ac": 0x40000,
                    "vif": 0x80000,
                    "vip": 0x100000,
                    "id": 0x200000,
                    "flags": 0xFFFF,
                    "eflags": 0xFFFFFFFF,
                }
            ),
        ]
        super(x86_64_Registers, self).__init__(registers)
        # Create separate collection dedicated to x87 FPU registers.
        self.__dict__["fpu"] = FPURegisters()

    def __deepcopy__(self, memo):
        copy = super(x86_64_Registers, self).__deepcopy__(memo)
        copy.__dict__["fpu"] = deepcopy(self.fpu, memo)
        return copy

    def __getattr__(self, reg_name):
        try:
            return super(x86_64_Registers, self).__getattr__(reg_name)
        except AttributeError:
            return self.fpu[reg_name]

    def __setattr__(self, reg_name, value):
        try:
            super(x86_64_Registers, self).__setattr__(reg_name, value)
        except AttributeError:
            self.fpu[reg_name] = value

    @property
    def names(self):
        # Overwriting names attribute so we can also provide fpu register names.
        return super(x86_64_Registers, self).names + self.fpu.names


class FPURegisters(RegisterMap):
    """
    Holds FPU related registers and handles stack manipulation.
    """

    _TAG_MAP = ["tag0", "tag1", "tag2", "tag3", "tag4", "tag5", "tag6", "tag7"]
    _ST_MAP = ["st0", "st1", "st2", "st3", "st4", "st5", "st6", "st7"]

    NaN = float("nan")
    INFINITY = float("inf")
    EMPTY = None

    def __init__(self):
        registers = [
            # control word
            Register(2, ic=0x1000, rc=0xC00, pc=0x300, iem=0x80, pm=0x20, um=0x10, om=0x8, zm=0x4, dm=0x2, im=0x1),
            # status word
            # TODO: IA-32 x87 indicates differently...
            Register(
                2,
                b=0x4000,
                c3=0x2000,
                top=0x1C00,
                c2=0x400,
                c1=0x200,
                c0=0x100,
                ir=0x80,
                sf=0x40,
                p=0x20,
                u=0x10,
                o=0x8,
                z=0x4,
                d=0x2,
                i=0x1,
            ),
            # tag word
            # NOTE: This register can't actually be accessed directly by the programmer but its
            # exposure is necessary for opcodes like FSTENV.
            # Therefore, these field names where made up by me and are not representative
            # of any real register names.
            Register(2, tag7=0xC000, tag6=0x3000, tag5=0xC00, tag4=0x300, tag3=0xC0, tag2=0x30, tag1=0xC, tag0=0x3),
        ]
        super(FPURegisters, self).__init__(registers)
        self.__dict__["_stack"] = [self.EMPTY] * 8

        # Initialize tag register with all empties.
        self.tag_word = 0xFFFF  # Set all stack slots to empty
        self.pc = 0b11  # 64 bit precision
        self.iem = 1  # interrupt mask disabled
        self.pm = 1  # precision mask
        self.um = 1  # underflow mask
        self.om = 1  # overflow mask
        self.zm = 1  # zero divide mask
        self.dm = 1  # denormalized operand mask
        self.im = 1  # invalid operation mask

    def __deepcopy__(self, memo):
        copy = super(FPURegisters, self).__deepcopy__(memo)
        copy.__dict__["_stack"] = list(self._stack)
        return copy

    def __getattr__(self, reg_name):
        # Handle special names.
        if reg_name == "control_word":
            return self._registers[0]._value
        elif reg_name == "status_word":
            return self._registers[1]._value
        elif reg_name == "tag_word":
            return self._registers[2]._value

        # "st*" register field is special and refers to the top of the stack.
        # TODO: Determine the name IDA uses to refer to the other stack elements.
        reg_name = reg_name.lower()
        if not reg_name.startswith("st"):
            return super(FPURegisters, self).__getattr__(reg_name)

        if reg_name == "st":
            index = 0
        else:
            try:
                index = self._ST_MAP.index(reg_name)
            except ValueError:
                raise AttributeError("Invalid register: {}".format(reg_name))
        return self._stack[(self.top + index) % 8]

    def __setattr__(self, reg_name, value):
        # Handle special names.
        if reg_name == "control_word":
            self._registers[0].__dict__["_value"] = value
            return
        elif reg_name == "status_word":
            self._registers[1].__dict__["_value"] = value
            return
        elif reg_name == "tag_word":
            self._registers[2].__dict__["_value"] = value
            return

        # "st" register field is special and refers to the top of the stack.
        reg_name = reg_name.lower()
        if not reg_name.startswith("st"):
            super(FPURegisters, self).__setattr__(reg_name, value)
            return

        if reg_name == "st":
            index = 0
        else:
            try:
                index = self._ST_MAP.index(reg_name)
            except ValueError:
                raise AttributeError("Invalid register: {}".format(reg_name))

        stack_index = (self.top + index) % 8
        self._stack[stack_index] = value

        # Also set the tag register.
        tag_field = self._TAG_MAP[stack_index]
        if value in (self.NaN, self.INFINITY):
            self[tag_field] = 0b10  # NaN, infinity, or denormal
        elif value != 0:
            self[tag_field] = 0b00  # non-zero flag
        elif value == 0:
            self[tag_field] = 0b01  # zero flag

    @property
    def names(self):
        # Overwrite to add dynamically built register fields.
        return super(FPURegisters, self).names + ["st"] + self._ST_MAP

    # TODO: Support setting sf register if value isn't free.

    def push(self, value):
        # Decrement stack pointer and then add value.
        self.top = (self.top - 1) % 8
        self.st = value
        logger.debug(":: Pushed on FPU stack: {}".format(value))

    def pop(self):
        # Retreive top of stack, mark it as empty, then increment stack pointer.
        ret = self.st
        self[self._TAG_MAP[self.top]] = 0b11  # mark as empty
        self.st = self.EMPTY
        self.top = (self.top + 1) % 8
        logger.debug(":: Popped off FPU stack: {}".format(ret))
        return ret



"""
Description: Extracts stack strings
AUTHOR: DC3
"""

from binascii import unhexlify

import idaapi
import idautils
import idc

import kordesii
from kordesii.utils import tracing

# main things to modify based on what you're looking for
STRING_GAP_TOLERANCE = 0  # should generally be 0
MAX_CHARACTER_WIDTH = 2  # should be 1 or 2
MIN_STR_LENGTH = 2  # do not make this less than 1
ALLOW_HEX = False

ASCII = True
MAX_CHARACTER_VALUE = 127 if ASCII else 2 ** (8 * MAX_CHARACTER_WIDTH) - 1

# FE
POS_FIRST = 0
POS_SECOND = 1
POS_THIRD = 2
POS_ALL = 3
POS_DATA = 10
POS_STACK = 20

IGNORED_MNEMONICS = ["cmp", "test"]

# fmt: off
JUMPS = [
    "ja", "jna",
    "jae", "jnae",
    "jb", "jnb",
    "jbe", "jnbe",
    "jc", "jnc",
    "jcxz", "jncxz",
    "jecxz", "jnecxz",
    "jrcxz", "jnrcxz",
    "je", "jne",
    "jg", "jne",
    "jge", "jnge",
    "jl", "jnl",
    "jle", "jnle",
    "jmp",
    "jmpe",
    "jmpf",
    "jo", "jno",
    "jp", "jnp",
    "jpe",
    "jpo",
    "js", "jns",
    "jz", "jnz",
]
# fmt: on


def get_function(ea):
    func = idaapi.func_t()
    func.start_ea = idc.get_func_attr(ea, idc.FUNCATTR_START)
    func.end_ea = idc.get_func_attr(ea, idc.FUNCATTR_END)
    func.flags = idc.get_func_attr(ea, idc.FUNCATTR_FLAGS)
    if idc.BADADDR == func.start_ea or idc.BADADDR == func.end_ea:
        return False
    else:
        return func


def get_functions():
    functions = {}
    for func_ea in idautils.Functions():
        fname = idaapi.get_func_name(func_ea)
        if fname and fname not in list(functions.keys()):
            func = get_function(func_ea)
            if func:
                functions[fname] = func
    return functions


def is_ascii(i):
    """
    Takes an int (supposedly) representing a char.
    """
    return i in [0, 9, 10, 13] or 31 < i < 127


def is_string_ascii(s):
    if isinstance(s, str):
        try:
            s.encode("ascii", errors="strict")
            return True
        except UnicodeEncodeError:
            return False
    for x in s:
        if not is_ascii(x):
            return False
    return True


class StackStrings(object):
    def __init__(self):
        self._strings = set()

    @property
    def strings(self):
        return self._strings

    def strings_in_range(self, start_ea, end_ea):
        strings = []
        for string in self.strings:
            if start_ea <= string.ea < end_ea:
                strings.append((string, "string"))
        return strings

    def get_char(self, char, char_width=1, allow_multi_chars=False, allow_hex=False):
        """
        Decode a character from char given the char_width.

        Returns decoded char, if found, otherwise BADADDR.
        """
        if isinstance(char, bytes):
            return char
        current_max_character_value = min(MAX_CHARACTER_VALUE, 2 ** (8 * char_width) - 1)
        if char == 9 or char == 10 or char == 13 or 31 < char < current_max_character_value:
            return str.encode(chr(char), "unicode-escape")
        elif char == 0:
            return char
        elif char > 2 * current_max_character_value and allow_multi_chars:
            formatted_char = format(int(char), "x").encode()  # format to hex, avoids 0x and L issues
            try:
                decoded = unhexlify(formatted_char)[::-1]  # decode if possible, and reverse it
                if len(decoded) > 1:
                    for thing in decoded:
                        thing = self.get_char(thing, char_width)
                        if not thing or thing == idc.BADADDR:
                            break
                    else:
                        return decoded
            except:
                pass
            try:
                decoded = formatted_char  # otherwise try to force to char
                if len(decoded) > 1:
                    for thing in decoded:
                        thing = self.get_char(thing, char_width)
                        if not thing or thing == idc.BADADDR:
                            break
                    else:
                        return decoded
            except:
                pass
        elif not isinstance(char, Filler) and allow_hex and char_width == 1:
            try:
                return bytes((char,))
            except:
                return idc.BADADDR
        return idc.BADADDR  # failed

    def consolidate_data_fragments(self, stack, continue_on_gap=True):
        """
         Given the stack dict, compresses it into a consolidated form. If continue_on_gap is true, it will be
         a list of chars with nulls terminating the distinct segments, otherwise it returns only the first
         distinct segment.
        """
        consolidated = []
        prev_key = min(stack.keys()) if stack else 0
        for key in sorted(stack.keys()):
            if key - prev_key > STRING_GAP_TOLERANCE + 1:
                if continue_on_gap:
                    consolidated.append((b"\x00", None))
                else:
                    return consolidated
            if stack[key][0] is not None and not isinstance(stack[key][0], Filler):
                consolidated.append(stack[key])
            prev_key = key
        return consolidated

    def extend_string_tuple(self, string_tuple, stack_tuple):
        string_tuple[0] += stack_tuple[0]
        string_tuple[1].append(stack_tuple[1])

    def parse_strings(self, stack, stack_min_value=None, stack_max_value=None, char_width=1):
        """ Attempts to parse strings out of stack. Returns list of string tuple information. """
        strs = []
        valid_key = lambda key: (stack_min_value is None or stack_min_value <= key) and (
            stack_max_value is None or key <= stack_max_value
        )
        stack = self.consolidate_data_fragments(
            {
                key: (self.get_char(value, char_width, allow_hex=ALLOW_HEX), ea)
                for key, (value, ea) in stack.items()
                if valid_key(key)
            },
            stack_min_value is None and stack_max_value is None,
        )
        temp_str = [b"", [], 0]
        current_width = 0
        require_zero = False
        if char_width == 2:
            require_zero = True
        found_zero = True
        for i in range(len(stack)):
            current_entry = stack[i]
            # if isinstance(current_entry[0], str):
            #     current_entry[0] = current_entry[0].encode("latin1")
            if isinstance(current_entry[0], int):
                is_null = current_entry[0] == 0
            elif isinstance(current_entry[0], bytes):
                is_null = all(c == 0 for c in current_entry[0])  # ignore all null strs
            else:
                is_null = bool(current_entry[0])
            if is_null:
                if current_width == char_width:
                    if len(temp_str[0]) >= MIN_STR_LENGTH:
                        strs.append(temp_str)
                    temp_str = [b"", [], 0]
                    current_width = 0
                elif current_width < char_width:
                    current_width += 1
                    found_zero = True
                else:
                    temp_str = [b"", [], 0]
            elif current_entry[0] == idc.BADADDR:
                if current_width < char_width:
                    current_width += 1
                    found_zero = True
                else:
                    if len(temp_str[0]) >= MIN_STR_LENGTH:
                        strs.append(temp_str)
                    current_width = 0
                    temp_str = [b"", [], 0]
            elif current_width == 0 or current_width == char_width:
                if require_zero:
                    if found_zero:
                        found_zero = False
                        self.extend_string_tuple(temp_str, current_entry)
                        temp_str[2] += char_width
                        current_width = 1
                    else:
                        if len(temp_str[0]) >= MIN_STR_LENGTH:
                            strs.append(temp_str)
                        temp_str = [b"", [], 0]
                        current_width = 0
                        found_zero = True
                else:
                    self.extend_string_tuple(temp_str, current_entry)
                    temp_str[2] += char_width
                    current_width = 1
            else:
                temp_str = [b"", [], 0]
                found_zero = False
                self.extend_string_tuple(temp_str, current_entry)
                temp_str[2] += char_width
                current_width = 1
        if len(temp_str[0]) >= MIN_STR_LENGTH and current_width == char_width:
            strs.append(temp_str)
            temp_str = [b"", [], 0]
        return strs

    def report_strings(self, strs, stack, stack_min_value=None, stack_max_value=None):
        """ Parses and returns Stack strings as Strings """
        for char_width in range(1, MAX_CHARACTER_WIDTH + 1):
            parsed_strs = self.parse_strings(stack, stack_min_value, stack_max_value, char_width)
            for string, eas, length in parsed_strs:
                start_ea, end_ea = self.find_start_and_end(eas)
                for string_obj in strs:
                    if string_obj[2] == string and string_obj[0] == start_ea and string_obj[1] == end_ea:
                        break
                else:  # if we didn't break
                    strs.add((start_ea, end_ea, string))
                    old_cmt = idc.get_cmt(eas[0], 0)
                    old_cmt = "" if not old_cmt else old_cmt
                    if not is_string_ascii(string):
                        new_cmt = string.hex()
                        new_cmt = old_cmt + "\nStack String (hex): " + new_cmt
                    else:
                        new_cmt = string.decode("unicode-escape").replace("\\x00", "\nStack String: ")
                        new_cmt = old_cmt + "\nStack String: " + new_cmt
                    new_cmt += "\nSize: " + str(length)
                    new_cmt = "\n".join(list(set(new_cmt.split("\n"))))  # Remove duplicates.
                    idc.set_cmt(eas[0], str(new_cmt).strip("\r\n"), 0)

    def find_start_and_end(self, eas):
        """Find the highest and lowest ea in the given list of eas"""
        eas = [ea for ea in eas if ea is not None]
        eas.sort()
        return eas[0], eas[-1]

    def clear_reg_if_needed(self, reg, regs):
        """Clears the state of a register, if it was known"""
        if reg and reg[0] in list(regs.keys()):
            del regs[reg[0]]

    def set_stack(self, offset, ea, pos, state):
        """
        Sets the stack dictionary, at the given offset, to contain the value at the given position at the given ea,
        performing a lookup in the register dictionary if needed.
        """
        fill = False
        if idc.get_operand_type(ea, pos) == idc.o_imm:
            val = idc.get_operand_value(ea, pos)
            state.stack[offset] = (val, ea)
            fill = True
        else:
            reg = tracing.get_reg_fam(tracing.get_opnd_replacement(ea, pos))
            if reg and reg[0] in state.regs:
                val = state.regs[reg[0]][0]
                state.stack[offset] = (state.regs[reg[0]][0], ea)
                fill = True
        if fill:
            for i in range(0, tracing.get_byte_size_of_operand(ea, pos)):
                state.stack[offset + i] = (val & 0xFF, ea)
                val //= 256

    def handle_lea(self, state):
        """Updates the state of the stack string finding based on an lea instruction"""
        if idc.get_operand_type(state.ea, POS_SECOND) == idc.o_reg:
            source_reg = tracing.get_reg_fam(tracing.get_opnd_replacement(state.ea, POS_SECOND))
            if source_reg and source_reg[0] in state.regs:
                value = state.regs[source_reg[0]]
            else:
                value = None
        else:
            value = tracing.get_operand_value_replacement(state.ea, POS_SECOND, state)
        if value is not None and value in state.stack:
            self.report_strings(state.strs, state.stack)
        self.clear_reg_if_needed(tracing.get_reg_fam(tracing.get_opnd_replacement(state.ea, POS_FIRST)), state.regs)

    def handle_call(self, state):
        """Updates the state of the stack string finding based on a call instruction"""
        stack_pointer = idc.get_spd(state.ea)
        next_ea = state.ea + idc.get_item_size(state.ea)
        stack_pointer_delta = idc.get_sp_delta(next_ea)
        if stack_pointer is not None and stack_pointer_delta is not None:
            next_reg = tracing.get_reg_fam(idc.print_operand(next_ea, POS_FIRST))
            # Caller cleanup handling, vulnerable to instruction reordering though.
            if next_reg and "esp" in next_reg and "add" in idc.print_insn_mnem(next_ea).lower():
                stack_pointer_delta += idc.get_sp_delta(next_ea + idc.get_item_size(next_ea))
            for index in range(stack_pointer, stack_pointer + stack_pointer_delta):
                if index in state.stack:
                    del state.stack[index]

    def handle_mov(self, state):
        """Updates the state of the stack string finding based on a mov instruction"""
        op1 = tracing.get_opnd_replacement(state.ea, POS_FIRST)
        if "[" in op1:
            offset = tracing.get_operand_value_replacement(state.ea, POS_FIRST, state)
            self.set_stack(offset, state.ea, POS_SECOND, state)
        else:
            reg = tracing.get_reg_fam(op1)
            type_ = idc.get_operand_type(state.ea, POS_SECOND)
            if reg:
                if type_ != idc.o_phrase and type_ != idc.o_displ:
                    if type_ == idc.o_reg:
                        reg2 = tracing.get_reg_fam(tracing.get_opnd_replacement(state.ea, POS_SECOND))
                        if reg2 and reg2[0] in state.regs:
                            val = state.regs[reg2[0]][0]
                        else:
                            val = None
                    else:
                        val = idc.get_operand_value(state.ea, POS_SECOND)
                    if val is not None:
                        state.regs[reg[0]] = (val, state.ea)
                else:
                    offset = tracing.get_operand_value_replacement(state.ea, POS_SECOND, state)
                    value = state.stack.get(offset, None)
                    if value is not None:
                        state.regs[reg[0]] = value
                    else:
                        self.clear_reg_if_needed(reg, state.regs)

    def get_stack_strings(self, functions):
        """
        Finds all the stack strings it can in the given functions.

        Parameters set globally:
            STRING_GAP_TOLERANCE - the gap allowed between string characters.
            MAX_CHARACTER_WIDTH  - the maximum character size, in bytes
            ASCII                - Whether character values must be 0-127
        """
        stack_strings = []
        for func in functions:
            state = tracing.BranchingTraceState(func.start_ea)
            state.strs = set()
            states = [state]
            func_eas = []
            ea = state.ea
            while ea < func.end_ea:
                func_eas.append(ea)
                ea += idc.get_item_size(ea)
            while states:
                state = states.pop()
                while state.ea < func.end_ea:
                    try:
                        func_eas.remove(state.ea)
                    except:
                        pass
                    state.visited_eas.append(state.ea)
                    mnemonic = idc.print_insn_mnem(state.ea)
                    if mnemonic in IGNORED_MNEMONICS:
                        pass
                    elif "pop" in mnemonic:
                        reg = tracing.get_reg_fam(tracing.get_opnd_replacement(state.ea, POS_FIRST))
                        if reg:
                            value = state.stack.get(idc.get_spd(state.ea), None)
                            if value is not None:
                                state.regs[reg[0]] = value
                            else:
                                self.clear_reg_if_needed(reg, state.regs)
                    elif "push" in mnemonic:
                        # bug where idc.get_spd was not correctly tracking the pointer,
                        # this case also hasn't really been seen often as part of a stack string
                        # self.set_stack(idc.get_spd(ea), ea, POS_FIRST, regs, stack)
                        pass
                    elif "mov" in mnemonic:
                        self.handle_mov(state)
                    elif (
                        (
                            "xor" in mnemonic
                            and tracing.get_reg_fam(tracing.get_opnd_replacement(state.ea, POS_FIRST))
                            == tracing.get_reg_fam(tracing.get_opnd_replacement(state.ea, POS_SECOND))
                        )
                        or ("lea" in mnemonic and idc.print_operand(state.ea, POS_SECOND) == "[0]")
                        or (
                            "sub" in mnemonic
                            and tracing.get_opnd_replacement(state.ea, POS_FIRST)
                            == tracing.get_opnd_replacement(state.ea, POS_SECOND)
                        )
                    ):
                        reg = tracing.get_reg_fam(tracing.get_opnd_replacement(state.ea, POS_FIRST))
                        if reg:
                            state.regs[reg[0]] = (0, state.ea)
                    elif "loop" in mnemonic or "movsb" in mnemonic:
                        state.regs["rcx"] = (0, state.ea)
                    elif mnemonic in JUMPS:
                        try:
                            target = next(idautils.CodeRefsFrom(state.ea, 0))
                        except StopIteration:
                            target = None
                        if target and target not in state.visited_eas:
                            if func.end_ea > target >= func.start_ea:
                                state.visited_eas.append(target)
                                new_state = tracing.BranchingTraceState(target, state)
                                new_state.strs = state.strs
                                states.append(new_state)
                            else:
                                self.report_strings(state.strs, state.stack)
                        # Always follow an unconditional jump
                        if mnemonic == "jmp":
                            break
                    elif (
                        "rep" in idc.GetDisasm(state.ea).split(" ")[0]
                        and "scas" not in idc.GetDisasm(state.ea).split(" ")[1]
                    ):
                        self.report_strings(state.strs, state.stack)
                    elif "lea" in mnemonic:
                        self.handle_lea(state)
                    elif "call" in mnemonic:
                        self.handle_call(state)
                    elif "ret" in mnemonic:
                        break
                    elif (
                        idc.get_operand_type(state.ea, POS_FIRST) == idc.o_reg
                    ):  # If we find a target register we were tracking, stop tracking it.
                        self.clear_reg_if_needed(
                            tracing.get_reg_fam(tracing.get_opnd_replacement(state.ea, POS_FIRST)), state.regs
                        )
                    state.ea += idc.get_item_size(state.ea)
                self.report_strings(state.strs, state.stack)
                if not states and func_eas:
                    new_state = tracing.BranchingTraceState(func_eas[0])
                    new_state.strs = set()
                    states.append(new_state)
                stack_strings.extend(state.strs)
        self.strings.update(stack_strings)


class Filler(object):
    """A class to represent the space that a wide char takes up during stack string finding"""

    def __new__(cls, *args, **kwargs):
        instance = cls.__dict__.get("__instance__", None)
        if instance is None:
            instance = super(Filler, cls).__new__(cls)
            instance.__init__()
            cls.__instance__ = instance
        return instance

    def __init__(self):
        super(Filler, self).__init__()

    def __bool__(self):
        return False

    def __int__(self):
        return 0

    def __long__(self):
        return self.__int__()


def getstackstrings():
    """
    Performs the work of retrieving the stack strings.

    :return: StackStrings object
    """
    stacked = StackStrings()
    functions = list(get_functions().values())
    stacked.get_stack_strings(functions)
    return stacked


@kordesii.decoder_entry
def main():
    stacked = getstackstrings()

    print("\n")
    for loc, end, string in sorted(stacked.strings, key=lambda tup: tup[0]):
        func_name: str = idaapi.get_func_name(loc)
        line = func_name.encode() + f", 0x{loc:X}: \x00".encode() + string + b'"'  # Use the \x00 for the replace below
        print((line.replace(b"\\x00", b'"\n\t\t"').replace(b'"\n', b"\n", 1) + b"\n").decode())
        kordesii.append_string(string.replace(b"\\x00", b"\n"))
    print("Found " + str(len(stacked.strings)) + " stack strings.")



from __future__ import absolute_import

import flask as f

from . import server


def create_app(extra_config=None):
    """
    Create a Flask app instance for the Kordesii API.

    :param dict extra_config: Extra configuration options to add to the app
    :return: Flask app for Kordesii API
    """
    app = f.Flask(__name__)

    app.config.setdefault("MENU_LINKS", []).extend(
        [
            {"name": "Upload", "endpoint": "kordesii.upload"},
            {"name": "Parsers", "endpoint": "kordesii.decoders_list"},
        ]
    )

    if extra_config:
        app.config.from_mapping(extra_config)

    server.init_app(app)
    app.register_blueprint(server.bp)

    return app



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""This is the setup file for the kordesii project."""
import sys
from setuptools import setup, find_packages


version_tuple = (sys.version_info[0], sys.version_info[1])

setup(
    name='kordesii',
    author='DC3',
    url='https://github.com/Defense-Cyber-Crime-Center/kordesii',
    keywords=['malware', 'ida', 'idapro'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    entry_points={
        'console_scripts': [
            'kordesii = kordesii.cli:main',
        ],
        'kordesii.decoders': [
            'kordesii = kordesii.decoders',
        ]
    },
    packages=find_packages(),
    include_package_data=True,
    license='MIT',
    python_requires='>=3.6',
    install_requires=[
        'click',
        'dill',
        'numpy',
        'pyelftools',
        'pefile>=2019.4.18',
        'pycryptodome',
        'requests',
        'Pyro4',
        'pyyaml',
        'tabulate',
        'yara-python',
        'ruamel.yaml',
        'setuptools',
        'six',

        # For the server and API
        'flask~=1.1.0',
        'pygments~=2.2.0',

        # Testing
        'pytest',
        'pytest-console-scripts',
        'tox',
    ]
)
